import re
from collections import defaultdict
import rdflib
from pyontutils.sheets import Sheet
from pyontutils.namespaces import ilxtr, TEMP, rdfs, skos, owl
from neurondm.core import Config, NeuronEBM, Phenotype, EntailedPhenotype, log, OntId
from neurondm.models.apinat_pops_more import _genlabel, NeuronApinatSimple, NeuronApinatComplex


class AtN(Sheet):
    name = 'apinat-to-npo-sue'


sheet_classes = [
    type(f'AtN{sname.replace(" ", "_")}',
         (AtN,), dict(sheet_name=sname))
    for sname in (
            'aacar',
            'splen',
            'sstom',
            'sdcol',

            'bromo',
            'kblad',
            'bolew',
            'pancr',
    )]


def map_predicates(sheet_pred):
    p = {
        '': TEMP.BROKEN_EMPTY,
        'Soma-Location': ilxtr.hasSomaLocatedIn,
        'Axon-Terminal-Location': ilxtr.hasAxonPresynapticElementIn,
        'Axon-Leading-To-Sensory-Terminal': ilxtr.hasAxonLeadingToSensorySubcellularElementIn,
        'Axon-Location': ilxtr.hasAxonLocatedIn,
        'Dendrite-Location': ilxtr.hasDendriteLocatedIn,
        #'Dendrite-Terminal-Location': ilxtr.hasDendriteSensorySubcellularElementIn,
        'Forward-Connection': ilxtr.hasForwardConnectionPhenotype,
        'Axon-Sensory-Location': ilxtr.hasAxonSensorySubcellularElementIn,
        'Equivalent-To': owl.equivalentClass,
        'Functional-Circuit-Role': ilxtr.hasFunctionalCircuitRolePhenotype,
        'entail:hasInstanceInTaxon': ilxtr.hasInstanceInTaxon,
        'Organ-Destination': ilxtr.hasOrganTarget,
        'Has-Phenotype': ilxtr.hasPhenotype,  # needed for aacar 7 and 8
        'Projection-Phenotype': ilxtr.hasProjectionPhenotype,
    }[sheet_pred]
    return p


def main(debug=False):
    Config('blank')  # XXX needed to avoid core_graph mixup with neurondm.core._g
    cs = [c() for c in sheet_classes]
    trips = [[cl] + [c.value for c in
                     (r.neuron_id(), r.exact_location(), r.location_id())]
             for cl in cs for r in cl.rows()
             if r.row_index > 0 and r.neuron_id().value
             and (not hasattr(r, 'exclude') or not r.exclude().value)]

    to_add = []
    dd = defaultdict(list)
    for c, _s, _p, _o in trips:
        for x in (_s, _p, _o):
            if re.match(r'(^[\s]+[^\s].*|.*[^\s][\s]+$)', x):
                msg = f'leading/trailing whitespace in {c} {_s!r} {_p!r} {_o!r}'
                log.error(msg)
                raise ValueError(msg)

        _s = _s.replace('/neuron-type-', '/apinat-simple/neuron-type-')
        s = OntId(_s)
        try:
            p = map_predicates(_p)
        except KeyError as e:
            log.error(f'sigh {s}')
            raise e

        if not _o:
            msg = f'object missing id for {OntId(s).curie} {OntId(p).curie}'
            if debug:
                log.warning(msg)
                _o = TEMP.BROKEN_EMPTY
            else:
                raise ValueError(msg)

        _o = _o.replace('/neuron-type-', '/apinat-simple/neuron-type-')  # needed at least for sstom-13 and sstom-14
        o = OntId(_o)

        if p == owl.equivalentClass or p == ilxtr.hasOrganTarget:
            to_add.append((s.u, p, o.u))
            continue

        try:
            if _p.startswith('entail:'):  # XXX FIXME hack
                dd[s].append(EntailedPhenotype(o, p))
            else:
                dd[s].append(Phenotype(o, p))
        except TypeError as e:
            log.error(f'bad data for {c} {s} {p} {o}')
            raise e

    problems = ('8a', '8v', 'sstom-6', 'keast-2', 'sdcol-k', 'sdcol-l')
    def eff(n):
        return bool([x for x in problems
                     if x in n.id_ and '20' not in n.id_])

    config = Config('apinat-simple-sheet')
    #import neurondm  # XXX alternate fix for core_graph _g mixup due to creating Phenotypes when Config has not been called
    #neurondm.core._g.namespace_manager.bind('apnsmp', str(ilxtr['apinat-simple/']))
    Phenotype.in_graph.namespace_manager.bind('apnsmp', str(ilxtr['apinat-simple/']))  # XXX needed along with Config('blank') above
    sigh = []
    nrns = []
    for id, phenos in dd.items():
        l, x = _genlabel(id)
        lab = 'simple ' + l
        to_add.append((id.u, rdfs.label, rdflib.Literal(lab)))
        n = NeuronApinatSimple(*phenos, id_=id)
        if eff(n):
            n._sigh()  # XXXXXXXXXXXXXXXXXXXXXXXXXXXXXXX FIXME figure out why this is not getting called internally
            sigh.append(n)

        nrns.append(n)

    config.write()
    labels = (
        ilxtr.genLabel, ilxtr.localLabel, ilxtr.simpleLabel,
        ilxtr.simpleLocalLabel, rdfs.label, skos.prefLabel)
    to_remove = [t for t in config._written_graph if t[1] in labels]
    [config._written_graph.remove(t) for t in to_remove]
    [config._written_graph.add(t) for t in to_add]
    config._written_graph.namespace_manager.bind('apnsmp', str(ilxtr['apinat-simple/']))
    config._written_graph.write()
    config.write_python()
    return config,


if __name__ == '__main__':
    main()



import os
import pathlib
import textwrap
from collections import defaultdict
import rdflib
import graphviz
from pyontutils.core import OntGraph, OntResIri, OntResPath
from pyontutils.namespaces import rdfs, ilxtr
from neurondm.core import Config, graphBase, log
from neurondm.core import OntTerm, OntId, RDFL
from neurondm.models.composer import load_config
from neurondm import orders


def multi_orig_dest(neuron):
    for dim in neuron.edges:
        if 'hasAxonPre' in dim or 'hasAxonSens' in dim or 'hasSoma' in dim:
            objs = list(neuron.getObjects(dim))
            if len(objs) > 1:
                return True


def makelpesrdf():
    collect = []
    def lpes(neuron, predicate):
        """ get predicates from python bags """
        # TODO could add expected cardinality here if needed
        return [(ph._neurdf_prefix_type, ph.p) for ph in neuron.pes
                if ph.e == predicate and not collect.append((ph._neurdf_prefix_type, predicate, ph.p))]

    def lrdf(neuron, predicate):
        """ get predicates from graph """
        return [  # XXX FIXME core_graph bad etc.
            str(o) for o in
            neuron.core_graph[neuron.identifier:predicate]]

    return lpes, lrdf, collect


def simplify(e):
    if e is None:
        return
    elif isinstance(e, rdflib.Literal):  # blank case
        return e.toPython()
    else:
        return OntTerm(e).curie


def simplify_nested(f, nested):
    if nested is None:
        return

    for e in nested:
        if isinstance(e, list) or isinstance(e, tuple):
            yield tuple(simplify_nested(f, e))
        elif isinstance(e, orders.rl):
            yield orders.rl(f(e.region), f(e.layer))
        else:
            yield f(e)


def filter_cycles(adj):
    out = []
    for a, b in adj:
        if a == b:
            log.warning(f'cycle removed on {a}')
            continue

        out.append((a, b))

    return out


from neurondm.models.apinat_pops_more import _genlabel
def wssn(nid):
    p, f = nid.rsplit('/', 1)
    if f.startswith('neuron-type-'):
        label, x = _genlabel(nid)
        return x
    elif 'sparc-nlp/' in nid:
        p, x, n = nid.rsplit('/', 2)
        return x
    else:
        return None


def for_composer(n, cull=False):
    lpes, lrdf, collect = makelpesrdf()
    _po = n.partialOrder()
    idwssn = wssn(n.id_)
    fc = dict(
        id = str(n.id_),
        label = str(n.origLabel),  # this is our rdfs label not prefLabel
        # prefLabel = ???,  # TODO I don't think we roundtrip this normally
        origin = lpes(n, ilxtr.hasSomaLocatedIn),
        dest = (
            # XXX looking at this there seems to be a fault assumption that
            # there is only a single destination type per statement, this is
            # not the case, there is destination type per destination
            [dict(loc=l, type='AXON-T') for l in lpes(n, ilxtr.hasAxonPresynapticElementIn)] +
            # XXX I strongly reccoment renaming this to SENSORY-T so that the
            # short forms are harder to confuse A-T and S-T
            [dict(loc=l, type='AFFERENT-T') for l in lpes(n, ilxtr.hasAxonSensorySubcellularElementIn)]
        ),
        order = tuple(simplify_nested(simplify, _po)) if _po else [],
        path = (  # TODO pull ordering from partial orders (not implemented in core atm)
            [dict(loc=l, type='AXON') for l in lpes(n, ilxtr.hasAxonLocatedIn)] +
            # XXX dendrites don't really ... via ... they are all both terminal and via at the same time ...
            [dict(loc=l, type='DENDRITE') for l in lpes(n, ilxtr.hasDendriteLocatedIn)]
        ),
        #laterality = lpes(n, ilxtr.hasLaterality),  # left/rigth tricky ?
        #projection_laterality = lpes(n, ilxtr.???),  # axon located in contra ?
        species =            lpes(n, ilxtr.hasInstanceInTaxon),
        sex =                lpes(n, ilxtr.hasBiologicalSex),
        circuit_type =       lpes(n, ilxtr.hasCircuitRolePhenotype),
        #phenotype =          lpes(n, ilxtr.hasAnatomicalSystemPhenotype),  # no longer current
        anatomical_system =  lpes(n, ilxtr.hasAnatomicalSystemPhenotype),
        # there are a number of dimensions that we aren't converting right now
        dont_know_fcrp =     lpes(n, ilxtr.hasFunctionalCircuitRolePhenotype),
        other_phenotype = (  lpes(n, ilxtr.hasPhenotype)
                           + lpes(n, ilxtr.hasMolecularPhenotype)
                           + lpes(n, ilxtr.hasProjectionPhenotype)),
        forward_connection = lpes(n, ilxtr.hasForwardConnectionPhenotype),

        # direct references from individual individual neurons
        provenance =      lrdf(n, ilxtr.literatureCitation),
        sentence_number = lrdf(n, ilxtr.sentenceNumber),
        note_alert =      lrdf(n, ilxtr.alertNote),
        working_set =     lrdf(n, ilxtr.inNLPWorkingSet),
        ws_short =        idwssn,  # lrdf(n, ilxtr.workingSetShortName),  # XXX TODO materialize probably

        # XXX provenance from ApiNATOMY models as a whole is not ingested
        # right now because composer lacks support for 1:n from neuron to
        # prov, (or rather lacks prov collections) and because it attaches
        # prov to the sentece, which does not exist for all neurons

        # TODO more ...
        # notes = ?

        # for _ignore, hasClassificationPhenotype is used for ApiNATOMY
        # unlikely to be encountered for real neurons any time soon
        _ignore = lpes(n, ilxtr.hasClassificationPhenotype),  # used to ensure we account for all phenotypes
    )
    npo = set((p._neurdf_prefix_type, p.e, p.p) for p in n.pes)
    cpo = set(collect)
    unaccounted_pos = npo - cpo
    if unaccounted_pos:
        log.warning(
            (n.id_, [[n.in_graph.namespace_manager.qname(e) for e in pos]
                     for npt, *pos in unaccounted_pos]))
    return {k:v for k, v in fc.items() if v} if cull else fc


def location_summary(neurons, anatent_simple=False):
    import csv
    locations = sorted(set(
        OntTerm(pe.p) for n in neurons for pe in n.pes
        if pe.e in n._location_predicates))
    [_.fetch() for _ in locations]

    _loc_src = sorted(set(
        (OntTerm(pe.p), n.id_) for n in neurons for pe in n.pes
        if pe.e in n._location_predicates))

    ls2 = dict()
    for _l, _s in _loc_src:
        if _l not in ls2:
            ls2[_l] = []

        ls2[_l].append(_s)

    loc_src = dict()
    for k, v in ls2.items():
        nlp = [iri for iri in v if '/sparc-nlp/' in iri]
        apinat = [iri for iri in v if '/neuron-type-' in iri]
        both = apinat and nlp
        source = 'both' if both else 'nlp' if nlp else 'apinat' if apinat else 'unknown'
        loc_src[k] = source

    def key(t):
        return (t.prefix, t.label[0].lower()
                if isinstance(t, tuple)
                else t.lower())

    if anatent_simple:
        header = 'label', 'curie', 'iri', 'source'
        rows = (
            [header] +
            [(_.label, _.curie, _.iri, loc_src[_]) for _ in sorted(locations, key=key)])
        with open('/tmp/npo-nlp-apinat-location-summary.csv', 'wt') as f:
            csv.writer(f, lineterminator='\n').writerows(rows)

        preds = sorted(set(e for n in neurons for e in n.edges))
        header = ['id'] + [OntId(p).curie for p in preds]
        _rows = [[n.id_, *[','.join(sorted([OntId(o).curie  for o in n.getObjects(p)])) for p in preds]] for n in neurons]
        rows = [header] + _rows
        with open('/tmp/npo-by-predicates.csv', 'wt') as f:
            csv.writer(f, lineterminator='\n').writerows(rows)

    else:
        header = 'o', 'o_label', 'o_synonym'
        rows = (
            [header] +
            [(_.iri, _.label, syn) for _ in sorted(locations, key=key)
             for syn in _.synonyms])
        with open('/tmp/anatomical_entities.csv', 'wt') as f:
            csv.writer(f, lineterminator='\n').writerows(rows)


def reconcile(n):
    lobjs = set(o for p in n._location_predicates._litmap.values() for o in n.getObjects(p))
    po_rl = set(e for pair in orders.nst_to_adj(n.partialOrder()) for e in pair)
    po_r = set(t.region if isinstance(t, orders.rl) else t for t in po_rl)
    po_l = set(t.layer for t in po_rl if isinstance(t, orders.rl))
    po_rl.difference_update({rdflib.Literal('blank')})
    po_r.difference_update({rdflib.Literal('blank')})
    po_l.difference_update({rdflib.Literal('blank')})
    #[if isinstance(e, orders.rl) else ]
    both = po_r & lobjs
    either = po_r | lobjs
    missing_axioms = po_r - lobjs
    missing_orders = lobjs - po_r
    missing_orders_rl =  lobjs - ( po_r | po_l )
    withl_missing_axioms = po_rl - lobjs
    withl_missing_orders = lobjs - po_rl
    ok_reg_l = not (missing_axioms or missing_orders_rl)
    ok_reg = not (missing_axioms or missing_orders)
    ok_rl = not (withl_missing_axioms or withl_missing_orders)
    return {
        'ok_reg': ok_reg,
        'ok_rl': ok_rl,
        'ok_reg_l': ok_reg_l,
        'withl_missing_axioms': withl_missing_axioms,
        'withl_missing_orders': withl_missing_orders,
        'missing_axioms': missing_axioms,
        'missing_orders': missing_orders,
        'missing_orders_rl': missing_orders_rl,
    }


def lgen(u):
    l = OntTerm(u).label
    if isinstance(l, tuple):
        log.warning(f'sigh {u} {l}')
        l = l[0]
    if l is None:
        c = OntId(u).curie
        log.warning(f'no label for {c}')
        return c,
    else:
        return textwrap.wrap(l, 30)


def synviz(neurons):
    index = {n.id_:n for n in neurons}
    idstrs = ('keast', 'aacar', 'sstom', 'pancr', 'bolew', 'sdcol', 'splen', 'bromo',  # XXX hack :/
              # FIXME mmset1 and 2 need ccomp to be easily visualized (ccomp is better for layout anyway)
              'semves', 'prostate', 'mmset1', 'mmset2', 'mmset4',)
    done = set()
    subsets = []
    for idstr in idstrs:
        subset = set(n for n in neurons if idstr in n.identifier)
        done.update(subset)
        subsets.append((subset, idstr))

    missed = set(neurons) - done
    if missed:
        subsets.append((missed, 'missed'))

    for subset, idstr in subsets:
        edges = [
            (pre, index[post])
            for pre in subset
            for post in pre.getObjects(ilxtr.hasForwardConnectionPhenotype)]

        dot = graphviz.Digraph(comment='forward connection')

        for ne in set(ne for nene in edges for ne in nene):
            dot.node(OntId(ne.id_).curie.replace(":", "-"),
                     label=OntId(ne.id_).curie,
                     #label=ne.origLabel
                     )

        for pre, post in edges:
            dot.edge(OntId(pre.id_).curie.replace(":", "-"), OntId(post.id_).curie.replace(":", "-"))

        base = pathlib.Path('/tmp/syntaptic/')
        if not base.exists():
            base.mkdir()

        dot.render(base / f'neurons-synaptic-connectivity-{idstr}', format='svg')


def gviz(n):
    nid = OntId(n.id_).curie
    dot = graphviz.Digraph(comment=nid)
    dot.attr('graph', labelloc='t')
    dot.attr('graph', label=n.origLabel)
    dot.attr('node', fontname='DejaVu Sans Mono')
    _bl = rdflib.Literal('blank')
    adj = [e for e in orders.nst_to_adj(n.partialOrder()) if e[0] != _bl]
    # FIXME labels need to be done once or it will be insanely slow
    nl = '\n'
    labels = {v: ((f'{v.hash_thing()}\n'
                   f'{nl.join(lgen(v.region))}\nx\n'
                   f'{nl.join(lgen(v.layer))}')
              if isinstance(v, orders.rl) else f'{OntId(v).curie}\n{nl.join(lgen(v))}')
              for v in set(v for e in adj for v in e)}
    for ab in adj:
        a, b = ab
        c, d = [n.hash_thing().replace(":", "-") if isinstance(n, orders.rl) else OntId(n).curie.replace(":", "-") for n in ab]
        dot.edge(c, d)
        # XXX dupes
        dot.node(c, label=labels[a])
        dot.node(d, label=labels[b])

    #print(dot.source)
    base = pathlib.Path('/tmp/partial-orders/')
    if not base.exists():
        base.mkdir()

    dot.render(base / f'neuron-po-{nid.replace(":", "-")}', format='svg')


def apinat_ontology_to_internal(blob):
    fixes = {
        # white matter issues, cns white matter layer is not a parent class of the spinal cord layer
        # the proliferation of terms for white matter in uberon without consistent modelling is a problem
        'UBERON:0016549': 'UBERON:0002318',
        # epicardium -> epicardial fat, issue is in the NPO representation I think? cause layer mismatch
        'UBERON:0002348': 'UBERON:0015129',
    }
    exclude = (
        # lingering old id not used anywhere (hopefully)
        'lyph-cardiovascular-heart-fat-layer',  # -> mat-tissue-cardiovascular-epicardial-fat
        'lyph-cardiovascular-heart-epicardium',  # related to the fix for 2348 -> 15129 above FIXME this won't work will it ... double replace
        #'mat-peritoneal-serosa-of-esophagus',  # XXX mismatch between wbcrm and sstom modelling I think?

    )
    out = {}
    lc = blob['localConventions']  # FIXME TODO use these to expand to OntId
    for d in (*blob['lyphs'], *blob['materials']):
        if 'ontologyTerms' not in d:
            log.warning(f'missing ontologyTerms: {d}')
            continue
        ots = d['ontologyTerms']
        id = d['id']
        if id in exclude:
            continue
        if len(ots) > 1 or not ots:
            log.error(ots)
            continue
        else:
            # FIXME use localConventions first
            ots0 = ots[0]
            if ots0 in fixes:
                log.info(f'replaced {ots0} -> {fixes[ots0]}')
                ots0 = fixes[ots0]

            oid = OntId(ots0)
            if oid.u in out:
                log.critical(f'DOUBLE ID FOR {oid}: {out[oid.u]} {id}')
                continue
            else:
                out[oid.u] = id

    return out


def npo_to_apinat(neurons):
    import requests
    wbrcm = "https://raw.githubusercontent.com/open-physiology/apinatomy-models/master/models/wbrcm/source/wbrcm.json"
    blob = requests.get(wbrcm).json()
    lookup = apinat_ontology_to_internal(blob)
    all_segments = [seggen(n, blob, lookup) for n in neurons]
    _ok_segs = [(i, segs) for i, segs in enumerate(all_segments) if segs and all([v[0] and v[1] for v_all in segs.values() for v in v_all])]
    globals().update(locals())
    ok = [i for i, s in _ok_segs]
    ok_segs = [s for i, s in _ok_segs]
    log.info(f'ok/all: {len(ok_segs)}/{len(all_segments)}')
    globals().update(locals())
    all_chains = [c for i, n in enumerate(neurons) for c in chaingen(n, blob, lookup, i)]
    ok_chains = [c for i, n in enumerate(neurons) if i in ok for c in chaingen(n, blob, lookup, i)]  # TODO group probably ...
    for c in all_chains:
        c['ok'] = c in ok_chains

    localConventions = {  # TODO derive from somewhere more sensible
        'too': 'https://apinatomy.org/uris/models/too-map/ids/',
        'wbkg': 'https://apinatomy.org/uris/models/wbrcm/ids/',
    }
    out = {
        'id': 'npo-neurons',
        'name': 'NPO neuron types as ApiNATOMY neuron populations',
        'imports': [wbrcm],
        'namespace': 'npo-apinat',
        'localConventions': [{'prefix': k, 'namespace': v} for k, v in localConventions.items()],
        'chains': all_chains,
     }
    return out


def get_layer_index(r_id, l_id, blob):
    things = [l for l in blob['lyphs'] + blob['materials'] if l['id'] == r_id]
    if not things:
        msg = f'sigh? {r_id}'
        raise ValueError(msg)

    l = things[0]
    if 'layers' in l:
        layers = l['layers']
        if l_id not in layers:
            log.error(f'sigh: {r_id} {l_id} {layers}')
        return layers.index(l_id)
    elif 'supertype' in l:
        return get_layer_index(l['supertype'], l_id, blob)
    else:
        raise ValueError('uhoh')


def seggen(n, blob, lookup):
    segments, dis = _seggen(n, blob, lookup)
    return segments


process_types = {
    ilxtr.hasAxonLocatedIn: 'wbkg:lt-axon-tube',
    ilxtr.hasDendriteLocatedIn: 'wbkg:lt-dend-bag',
    ilxtr.hasSomaLocatedIn: ('wbkg:lt-soma-of-neuron', 'wbkg:lt-axon-tube'),  # FIXME see what to do about inferred dendrites
    # FIXME also issue with inferring dendrite bag because we don't know whether the might be an explicit dendrite
    ilxtr.hasAxonPresynapticElementIn: 'wbkg:lt-axon-bag',
    ilxtr.hasAxonSensorySubcellularElementIn: 'lt-axon-sens-bag',
    ilxtr.hasAxonLeadingToSensorySubcellularElementIn: 'lt-axon-sens-tube',
}


def _seggen(n, blob, lookup):
    _blank = rdflib.Literal('blank')
    adj = [edge for edge in orders.nst_to_adj(n.partialOrder()) if edge[0] != _blank]
    try:
        dis, lin = orders.adj_to_lin(adj)
    except Exception as e:
        log.exception(e)
        return {}, None

    starts = [seq[0] for seq in dis]
    int_dis = [seq[-1] for seq in dis if len(seq) > 1 and seq[-1] in starts]
    assert not int_dis, 'derp'
    in_linkers = set(v for edge in lin for v in edge)
    tar_linkers = {}
    src_linkers = {}
    dd = defaultdict(set)
    idp = OntId(n.identifier).curie.replace(":", "-")
    for i, edge in enumerate(lin):
        node_id = f'node-{idp}-{i}'
        a, b = edge
        if a in tar_linkers and b in src_linkers and tar_linkers[a] != src_linkers[b]:
            # this is the case where we have to rewrite nodes because we discover that
            # they converge at some later point because a vertex is present in multiple
            # linkers, in this case we prefer and thus use the lowest existing node id
            tla = tar_linkers[a]
            slb = src_linkers[b]
            node_id, rewrite = (tla, slb) if tla < slb else (slb, tla)
            to_rewrite = dd[rewrite]
            dd[rewrite] = None
            for identifier in to_rewrite:
                for _linkers in (tar_linkers, src_linkers):
                    if identifier in _linkers and _linkers[identifier] == rewrite:
                        # must also check that the identifier matches rewrite otherwise we overwrite
                        # a regions target with its source or source with its target
                        _linkers[identifier] = node_id

        elif a in tar_linkers:
            node_id = tar_linkers[a]
        elif b in src_linkers:
            node_id = src_linkers[b]

        tar_linkers[a] = node_id
        src_linkers[b] = node_id
        dd[node_id].update((a, b))

    lindex = dict(dd)

    # axiom lookup
    alu = {o:p for p in n._location_predicates._litmap.values() for o in n.getObjects(p)}
    po_rl = set(v for edge in adj for v in edge)

    segments = {}
    for vertex in po_rl:
        all_segs = []
        seg = []
        for e in ((vertex.region, vertex.layer) if isinstance(vertex, orders.rl) else (vertex, None)):
            if e in lookup:
                i = lookup[e]
            else:
                i = None

            if e in alu:
                t = alu[e]
            else:
                t = None

            seg.append(i)
            seg.append(t)

            if e in tar_linkers:
                ta = tar_linkers[e]
            else:
                ta = None

            if e in src_linkers:
                sr = src_linkers[e]
            else:
                sr = None

            seg.append(ta)
            seg.append(sr)

        r_id, r_type, r_tar, r_src, l_id, l_type, l_tar, l_src = seg
        if r_type and l_type and r_type != l_type:
            msg = f'error rendring {n}: {r_type} != {l_type}'
            log.error(msg)
            return {}, dis
        elif r_type:
            # FIXME soma -> + axon
            actual_type = process_types[r_type]
        elif l_type:
            log.critical(f'uhoh {r_id} {l_id} {l_type}')
            actual_type = process_types[l_type]
        else:
            actual_type = None

        try:
            layer_index = None if r_id is None or l_id is None else get_layer_index(r_id, l_id, blob)
        except ValueError as e:
            log.exception(e)
            msg = f'error rendring {n}'
            log.error(msg)
            return {}, dis

        if isinstance(actual_type, tuple):
            # soma implied axon case
            if len(actual_type) < 2:
                raise ValueError('need at least two')

            at0, *atd, atn = actual_type
            s0 = r_id, r_type, None, r_src, l_id, l_type, None, l_src, at0, layer_index
            all_segs.append(s0)
            for at in atd:
                nseg = r_id, r_type, None, None, l_id, l_type, None, None, at, layer_index
                all_segs.append(nseg)

            sn = r_id, r_type, r_tar, None, l_id, l_type, l_tar, None, atn, layer_index
            all_segs.append(sn)
        else:
            seg.append(actual_type)
            seg.append(layer_index)
            all_segs.append(seg)

        segments[vertex] = all_segs  # FIXME needs to be a list because a single soma segment can expand into multiple chain segments

    #if int_dis or src_linkers or tar_linkers:
        #breakpoint()

    return segments, dis


def chaingen(n, blob, lookup, index=0):
    segments, dis = _seggen(n, blob, lookup)
    if not segments:
        return []

    chains = []
    for i, seq in enumerate(dis):
        lyphTemplates = []
        housingLyphs = []
        housingLayers = []
        levels = []
        root = None
        #leaf = None
        for j, vertex in enumerate(seq):  # XXX FIXME have to split chains by type because they can only have a single template :/
            all_segs = segments[vertex]
            for k, seg in enumerate(all_segs):
                r_id, r_type, r_tar, r_src, l_id, l_type, l_tar, l_src, actual_type, layer_index = seg

                lyphTemplates.append(actual_type)
                housingLyphs.append(None if r_id is None else 'wbkg:' + r_id)
                housingLayers.append(layer_index)  # XXX looking at housingLayers in pancreas it is clearly wrong

                # either nil or empty object if we don't need any explicit information
                link = {}

                if j == 0 and r_src and k == 0:
                    root = r_src  # FIXME there is no way there are 8 that share a root for mmset1/11
                elif r_src:
                    # XXX shouldn't need this even in multiply branched cases
                    # because it should be matched correctly by apinatomy ???
                    # but there could be a bug in the chaingen code for levels?
                    #link['source'] = r_src
                    #breakpoint()
                    pass

                if actual_type is not None:
                    link['conveyingLyph'] = actual_type #'lyph-template-type',
                if r_tar is not None:
                    # FIXME mmset1-11 everything has the same target, the 0th node which is also the root ???
                    link['target'] = r_tar

                if link:
                    # FIXME seems like there is a bug where if the link structure is missing and id
                    # the expansion code will incorrectly match it with other identical link objects
                    # resulting in links appearing to be shared between many neurons
                    # adding the id here is a workaround
                    link['id'] = f'link-for-chains-{index}-{i}-{j}-{k}'

                levels.append(link)  # TODO vs None

        chain = {
            'id': f'chain-{OntId(n.id_).curie.replace(":", "-")}-{i}',
            #'leaf': '',
            # 'lyphTemplates': lyphTemplates,  # not useful for branched structures
            'housingLyphs': housingLyphs,
            'housingLayers': housingLayers,
            'levels': levels,  # these are links so have to provide {'id': 'link-id', 'source': }
         }

        if root is not None:
            chain['root'] = root

        chains.append(chain)

    return chains


def main(local=False, anatomical_entities=False, anatent_simple=False, do_reconcile=False, viz=False, chains=False):
    # if (local := True, anatomical_entities := True, anatent_simple := False, do_reconcile := False, viz := False, chains := False):
    restore = do_reconcile or viz or anatomical_entities
    neurons, ex_config, ex_g = load_config(local=local, restore=restore)
    new_neurons = [n for n in neurons if 'neuron-type-' not in n.id_]  # filter out existing apinatomy populations

    # ingest to composer starts here
    mvp_ingest = [n for n in neurons if not multi_orig_dest(n)]

    dims = set(p for n in neurons for p in n.edges)  # for reference
    fcs = [for_composer(n) for n in mvp_ingest]
    _fcne = [for_composer(n, cull=True) for n in mvp_ingest]  # exclude empties for easier manual review

    # example neuron
    n = mvp_ingest[0]
    fc = for_composer(n)

    # example partial orders
    view_partial_orders = [tuple(simplify_nested(simplify, n.partialOrder())) for n in neurons]

    # example linearized orders
    def linearize(n):
        def key(d):
            return tuple((n.region, n.layer) if isinstance(n, orders.rl) else (n, '') for n in d)
            # FIXME somehow the caste to orders.rl here results in a different sort order than
            # the version above !?!??!
            #return tuple(n if isinstance(n, orders.rl) else orders.rl(n) for n in d)

        try:
            adj = filter_cycles(orders.nst_to_adj(tuple(simplify_nested(simplify, n.partialOrder()))))
            dis, lin = orders.adj_to_lin(adj)
            rej = orders.lin_to_adj(dis, lin)
            sa, sj = sorted(adj, key=key), sorted(rej, key=key)
            assert set(sa) == set(sj), breakpoint()  # XXX if this passes be the below fails we are in big trouble
            assert sa == sj, breakpoint()
            return dis, lin
        except Exception as e:
            log.exception(e)
            log.error(f'cycle in\n{n}')
            return [], []

    def _rend(dislin):
        dis, lin = dislin
        return [' '.join([n.hash_thing() if isinstance(n, orders.rl) else n for n in d])
                for d in sorted(dis, key=len, reverse=True)], [[n.hash_thing() if isinstance(n, orders.rl) else n for n in d] for d in lin]

    linear_orders = sorted([(n.id_, linearize(n)) for n in neurons])
    view_linear_orders = [(i, _rend(dislin)) for i, dislin in linear_orders]

    if do_reconcile:
        _recs = [(n, reconcile(n)) for n in neurons]
        recs_reg = [(n, r) for n, r in _recs if not r['ok_reg']]
        recs_rl = [(n, r) for n, r in _recs if not r['ok_rl']]
        recs_reg_l = [(n, r) for n, r in _recs if not r['ok_reg_l']]
        msg = f'{len(recs_reg)} pops with reg issues, {len(recs_rl)} pops with rl issues, {len(recs_reg_l)} pops with reg_l'
        log.info(msg)
        sigh_reg = sorted([(len(r["missing_axioms"]), len(r["missing_orders"]), n, r) for n, r in recs_reg],
                          key=lambda t: (t[0] + t[1], t[0], t[1]), reverse=True)
        sigh_how = [s[:2] + tuple(OntId(_.id_).curie for _ in s[2:3]) for s in sigh_reg]
        rep_reg = 'a  o  i\n' + '\n'.join(f'{a: >2} {o: >2} {i}' for a, o, i in sigh_how)
        missing_orders = sorted(set([e for n, r in _recs for e in r['missing_orders']]))
        missing_orders_rl = sorted(set([e for n, r in _recs for e in r['missing_orders_rl']]))

        # missing orders populations where neither region nor layer are accounted for
        morl = [(n, r) for n, r in _recs if r['missing_orders_rl']]

        import pprint
        from pyontutils.core import IlxTerm
        from neurondm.orders import rl
        from collections import Counter
        derp = defaultdict(list)
        for n, r in _recs:
            if r['missing_orders_rl']:
                for mo in r['missing_orders_rl']:
                    derp[mo].append(n)
        dderp = {OntTerm(k): v for k, v in derp.items()}
        mcmorl_issues = Counter([_.id_ for k, v in dderp.items() for _ in v]).most_common()

        derps = set([_ for k, v in dderp.items() for _ in v])
        lderps = len(derps)  # 63 for mo, 25 for

        _report = ''
        for k, v in dderp.items():
            _report += '\n--------------------------------------\n'
            _report += repr(k)
            for _ in v:
                _report += '\n'
                _report += str(_)
                _report += pprint.pformat(_.partialOrder())

        log.debug(_report)

        lu_mo = [OntTerm(t) for t in missing_orders]
        lu_mo_rl = [OntTerm(t) for t in missing_orders_rl]

        ilu_mo = [IlxTerm(t) for t in missing_orders]
        ilu_mo_rl = [IlxTerm(t) for t in missing_orders_rl]

        def sigh(t, pred):
            if not t.validated:
                # FIXME this is where query services wants RDFL
                # for things like ilxtr:cardiac-interganglionic-nerve
                # but the bug is really in OntTerm which should still
                # be able to answer predicates when not validated ...
                return tuple()

            if pred in t.predicates:
                v = t.predicates[pred]
                if not isinstance(v, tuple):
                    return v,

                return v

            return tuple()

        # XXX cases where region/layer were modelled as subClassOf/partOf pairs  # XXX probably need to fix modeling of these in interlex
        ilu_scopo = [(t, sigh(t, 'rdfs:subClassOf'), sigh(t, 'ilx.partOf:')) for t in ilu_mo_rl]
        ilu = [(a, rl(b[0].u, c[0].u)) for a, b, c in ilu_scopo if b and c if len(b) == 1 and len(c) == 1]

        _manual_lookup_table = {
            '': rl('', ''),
        }

        # XXX cases where region/layer/layer were modelled as part part vessles in layer of colon mostly
        # sdcol-f is not actually the issue here it seems
        #f = [n for n in neurons if 'sdcol-f' in n.id_][0]
        #f.pes()
        #f.partialOrder()

        # XXX cases where an axiom appears as a layer in a partial order because the layer is already a direct part of the region e.g. bromo-1 case
        # TODO try to automate detection of these cases during conversion from apinatomy to partial orders
        #b = [n for n in neurons if 'bromo-1' in n.id_][0]
        #b.pes()
        #b.partialOrder()

        # XXX cases where a super class was used (e.g. toracic instead of T1-T5)
        # TODO could try to add a subClassOf check if other matches fail? but doesn't resolve the desire for an exact match
        # e.g. bromo-1

        # XXX cases where the partial order split a structure into two parts e.g. proximal and distal stomach but axioms use the parent part
        # e.g. sstom-5

        #reg_missing_orders = [r['missing_orders'] for n, r in recs_reg]
        #rl_missing_orders = [r['missing_orders'] for n, r in recs_rl]
        breakpoint()

    if anatomical_entities:
        location_summary(neurons, anatent_simple)

    if viz:
        synviz(neurons)
        [gviz(n) for n in neurons]

    if chains:
        import json
        chains = npo_to_apinat(neurons)
        #chains = npo_to_apinat([n for n in neurons if 'mmset1/11' in n.id_])
        ok = [c for c in chains['chains'] if c['ok']]
        wl = [c for c in chains['chains'] if [_ for _ in c['levels'] if _]]
        wr = [c for c in chains['chains'] if 'root' in c and c['root']]
        wlr = [c for c in chains['chains'] if 'root' in c and c['root'] and [_ for _ in c['levels'] if _]]
        with open('/tmp/test-npo-apinat-chains.json', 'wt') as f:
            json.dump(chains, f, indent=2)


if __name__ == '__main__':
    main()



import re
from collections import defaultdict
from urllib.parse import quote as url_quote
import rdflib
from pyontutils.sheets import Sheet
from pyontutils.namespaces import ilxtr, TEMP, rdfs, skos, owl, interlex_namespace
from neurondm.core import Config, NeuronEBM, Phenotype, EntailedPhenotype, NegPhenotype, LogicalPhenotype, log, OntCuries, OntId, add_partial_orders, IntersectionOf, AND, OR
from neurondm import orders


class NeuronSparcNlp(NeuronEBM):
    owlClass = ilxtr.NeuronSparcNlp
    shortname = 'sprcnlp'
    _no_origLabel_hack = True


class NLP1(Sheet):
    name = 'off-nlp-1'


class NLP2(Sheet):
    name = 'off-nlp-2'


class NLPSemVes(Sheet):
    name = 'off-nlp-il-semves'


class NLPProst(Sheet):
    name = 'off-nlp-il-prostate'


class NLPFemrep(Sheet):
    name = 'off-nlp-il-femrep'


class NLPFemreprat(NLPFemrep):
    pass


class NLPSenseMotor(Sheet):
    name = 'sensory-motor'


class NLPLiver(Sheet):
    name = 'off-nlp-liv'


class NLPKidney(Sheet):
    name = 'off-nlp-kid'


class NLPSwglnd(Sheet):
    name = 'off-nlp-swglnd'


def nlp_ns(name):
    return rdflib.Namespace(interlex_namespace(f'tgbugs/uris/readable/sparc-nlp/{name}/'))


snames = {
    'MM Set 1': (NLP1, nlp_ns('mmset1'), None),
    'MM Set 2 Cranial Nerves': (NLP1, nlp_ns('mmset2cn'), 'cranial nerves'),
    'MM Set 4': (NLP2, nlp_ns('mmset4'), None),
    'Seminal Vesicles': (NLPSemVes, nlp_ns('semves'), 'seminal vesicles'),
    'Prostate': (NLPProst, nlp_ns('prostate'), 'prostate'),
    '1. Female Reproductive-HUMAN': (NLPFemrep, nlp_ns('femrep'), 'female reproductive system'),
    '3. Female-RAT': (NLPFemreprat, nlp_ns('femrep'), 'female reproductive system'),  # separate sheet ids differ
    'All SM connections': (NLPSenseMotor, nlp_ns('senmot'), 'sensory motor'),
    'ALL Liver_Human_Rat_Mouse': (NLPLiver, nlp_ns('liver'), 'liver'),
    'All KIDNEY connections': (NLPKidney, nlp_ns('kidney'), 'kidney'),
    'Sheet1': (NLPSwglnd, nlp_ns('swglnd'), 'sweat glands'),
    'comprt': (type('ComposerRT', (object,), dict()), None, 'composer round trip'),
    'NPO-Template (Individual cell types/species)': (type('Bhuiyan2024', (object,), dict()), None, None),
}


sheet_classes = [
    type(f'{base.__name__}{sname.replace(" ", "_")}',
         (base,), dict(sheet_name=sname))
    for sname, (base, ns, working_set) in snames.items()
    if ns is not None]


def make_annotation_properties(prefix=ilxtr):
    return (
        prefix.hasComposerURI,
        prefix.alertNote,
        prefix.curatorNote,
        prefix.expertConsultant,
        prefix.literatureCitation,
        rdflib.URIRef('http://uri.interlex.org/composer/uris/readable/hasComposerURI'),
        rdflib.URIRef('http://uri.interlex.org/tgbugs/uris/readable/composerGenLabel'),
    )


def map_uris(sheet_uri, prefix=ilxtr):
    try:
        return {
            ilxtr.hasProjection: prefix.hasProjectionPhenotype,  # FIXME iri is not real
            rdflib.URIRef('http://uri.interlex.org/tgbugs/uris/readable/neurdf/pred/ent/hasMolecularPhenotype'): prefix.hasMolecularPhenotype,

        }[sheet_uri]
    except KeyError as e:
        return sheet_uri


def map_predicates(sheet_pred, prefix=ilxtr):  # FIXME use the closed namespace
    p = {
        '': TEMP.BROKEN_EMPTY,
        'Soma': prefix.hasSomaLocatedIn,
        'Axon terminal': prefix.hasAxonPresynapticElementIn,
        'Axon': prefix.hasAxonLocatedIn,
        'Dendrite': prefix.hasDendriteLocatedIn,
        'Axon sensory terminal': prefix.hasAxonSensorySubcellularElementIn,
        'hasInstanceInTaxon': prefix.hasInstanceInTaxon,
        'hasPhenotype': prefix.hasPhenotype,
        'hasAnatomicalSystemPhenotype': prefix.hasAnatomicalSystemPhenotype,
        'hasBiologicalSex': prefix.hasBiologicalSex,
        'hasCircuitRole': prefix.hasCircuitRolePhenotype,  # used in femrep
        'hasCircuitRolePhenotype': prefix.hasCircuitRolePhenotype,
        'hasFunctionalCircuitRolePhenotype': prefix.hasFunctionalCircuitRolePhenotype,
        'hasForwardConnectionPhenotype': prefix.hasForwardConnectionPhenotype,  # FIXME this needs to be unionOf
        'Axon-Leading-To-Sensory-Terminal': prefix.hasAxonLeadingToSensorySubcellularElementIn,
        'hasAxonLeadingToSensorySubcellularElementIn': prefix.hasAxonLeadingToSensorySubcellularElementIn,

        # from composer
        'hasSomaLocatedIn': prefix.hasSomaLocatedIn,
        'hasDendriteLocatedIn': prefix.hasDendriteLocatedIn,
        'hasAxonLocatedIn': prefix.hasAxonLocatedIn,
        'hasAxonPresynapticElementIn': prefix.hasAxonPresynapticElementIn,
        'hasAxonSensorySubcellularElementIn': prefix.hasAxonSensorySubcellularElementIn,
        'hasProjection': prefix.hasProjectionPhenotype,  # XXX check the semantics on this one
        'hasProjectionPhenotype': prefix.hasProjectionPhenotype,  # XXX check the semantics on this one
        'hasProjectionLaterality': prefix.hasProjectionLaterality,  # XXX check the semantics on this one because it expects contra/ipsi not left/right
        'hasSomaPhenotype': prefix.hasSomaPhenotype,  # XXX what is this being used for? I can't find any objects?
        'hasComposerURI': prefix.hasComposerURI,  # annotation property
        'alertNote': prefix.alertNote,  # FIXME we have predicate uri in composer now
    }[sheet_pred]
    return p


def ind_to_adj(ind_uri, *, neuron_id=None):
    inds = sorted(set([i for i, u in ind_uri]))
    edges = []
    for a, b in zip(inds[:-1], inds[1:]):
        for i, iu in ind_uri:
            # XXX yes this is a horribly inefficient way to do this
            # index them in a dict if it becomes an issue
            if i == a:
                for j, ju in ind_uri:
                    if j == b:
                        edge =  iu, ju
                        edges.append(edge)

    sedges = set(edges)
    if len(edges) != len(sedges):
        log.warning(f'duplicate edges detected for {neuron_id}')

    return sorted(sedges)


def main(debug=False, cs=None, config=None, neuron_class=None, neuron_class_fun=None):
    def derp(v):
        class sigh:
            value = v
        return sigh

    def procobju(text, predicate=None):
        # TODO consider whether we can handle predicate here ...
        # despite that this is all a horrible hack
        if ' ' not in text:
            return OntId(text)
        else:
            log.debug(repr(text))
            uri, level = text.split(' ')
            if level == 'low':
                qual = ilxtr.LowerExpression
            elif level == 'high':
                qual = ilxtr.HigherExpression
            else:
                raise NotImplementedError(level)

            # FIXME hardcoded and assumes predicate
            return LogicalPhenotype(AND,
                                    Phenotype(OntId(uri),
                                              ilxtr.hasExpressionPhenotype),
                                    Phenotype(qual, ilxtr.hasPhenotypeModifier))

    OntCuries({'ISBN13': 'https://uilx.org/tgbugs/u/r/isbn-13/',})
    csin = cs
    if cs is None:
        cs = [c() for c in sheet_classes]

    annotation_properties = make_annotation_properties()

    trips = [[cl, (r.union_set().value.strip() if hasattr(r, 'union_set') else None)] + [
        c if isinstance(c, OntId) or isinstance(c, LogicalPhenotype) else c.value for c in
        ((r.id() if hasattr(r, 'id') else OntId(r.subject_uri().value)),
         (OntId(r.predicate_uri().value) if hasattr(r, 'predicate_uri') else
          (r.predicate() if hasattr(r, 'predicate') else r.relationship())),
         ((r.identifier() if r.identifier().value.strip()
           else derp(TEMP['MISSING_' + r.structure().value.replace(' ', '-')]))
          if hasattr(r, 'identifier') else
          (procobju(r.object_uri().value) if r.object_uri().value.strip()  # FIXME need to fill object text case, likely below
           else derp(TEMP['MISSING_' + r.object().value.replace(' ', '-')]))))]
             for cl in cs for r in cl.rows()
             if r.row_index > 0 and (r.id().value if hasattr(r, 'id') else r.subject_uri().value)
             #and (not hasattr(r, 'exclude') or not r.exclude().value)
             and r.proposed_action().value.lower() != "don't add"
             # FIXME we handle annotation properties separately right now
             # all this code is a mess so just roll with it for now
             and not (hasattr(r, 'object_text') and r.object_text().value.strip())
             #and not log.debug(r.neuron_id())
             ]

    to_add = []
    def to_add_append(t):
        s, p, o = t
        _t = tuple([e.u if isinstance(e, OntId) else e for e in t])
        to_add.append(_t)

    def vl(meth):
        c = meth()
        v = c.value
        vs = v.strip()
        if v != vs:
            msg = f'whitespace issue in {c}'
            log.warning(msg)

        return rdflib.Literal(vs)

    extra_kwargs = {}
    def asdf(s, p, rm, split=False, rdf_type=rdflib.Literal, as_kwarg=None):
        vin = vl(rm)
        if vin:
            # XXX watch out for non-homogenous subject types
            # (e.g. OntId vs rdflib.URIRef)
            if split:
                for _v in vin.split(split):
                    v = _v.strip()
                    if ' ' in v:
                        log.error((s, p, v))
                        v = v.replace(' ', '')

                    if (v.startswith('https:/doi.org') or v.startswith('http:/w') or
                        v.startswith('https:/git') or v.startswith('http:/uri.interlex.org')):
                        log.error((s, p, v))
                        v = v.replace(':/', '://', 1)
                    try:
                        o = rdf_type(v)
                        to_add_append((s.u, p, o))
                    except Exception as e:
                        breakpoint()
                        raise e
            else:
                if as_kwarg:
                    if s not in extra_kwargs:
                        extra_kwargs[s] = {}

                    extra_kwargs[s][as_kwarg] = vin
                else:
                    to_add_append((s.u, p, vin))

    def bind_lcc(c):
        def lcc(uri_or_curie, _c=c):
            if '<' in uri_or_curie:
                # see pyontutils.utils_extra check_value
                return rdflib.URIRef(url_quote(uri_or_curie, ':/;()'))
            elif uri_or_curie.startswith('10.'):
                log.error(f'bad doi {uri_or_curie} in {_c}')
                return OntId('DOI:' + uri_or_curie).u
            else:
                return OntId(uri_or_curie).u

        return lcc

    dd = defaultdict(list)
    cfdd = defaultdict(list)
    ec = {}
    def wrap(s):
        class cl:
            value = s
        def inner(_=cl):
            return _

        return inner

    for cl in cs:
        _, nlpns, working_set = snames[cl.sheet_name]
        log.debug(f'working on {working_set}')
        for r in cl.rows():
            try:
                if (r.row_index > 0 and
                    (r.id if hasattr(r, 'id') else r.subject_uri)().value and
                    r.proposed_action().value.lower() != "don't add"):
                    # extra trips
                    #print(repr(r.id()))
                    _id = (r.id().value if hasattr(r, 'id') else OntId(r.subject_uri().value))
                    _prefix = nlpns.split('/')[-2] if hasattr(r, 'id') else 'comprt'  # FIXME cmpsr
                    s = _id if isinstance(_id, OntId) else OntId(nlpns[_id])
                    #log.debug(s)
                    if hasattr(r, 'sentence_number'):
                        asdf(s, ilxtr.sentenceNumber, r.sentence_number)
                    if hasattr(r, 'different_from_existing'):
                        asdf(s, ilxtr.curatorNote, r.different_from_existing)
                    if csin is None:  # hack to avoid the mapping in composer
                        asdf(s, ilxtr.curatorNote, r.curation_notes)
                    else:
                        # FIXME bad assumption that csin is always composer
                        if hasattr(r, 'curation_notes'):
                            asdf(s, ilxtr.systemNote, r.curation_notes)

                    if hasattr(r, 'review_notes'):
                        asdf(s, ilxtr.reviewNote, r.review_notes)

                    if hasattr(r, 'reference_pubmed_id__doi_or_text'):
                        # meaning change in composer to literature_citation
                        # XXX TODO but that means literature_citation must handle freetext
                        asdf(s, ilxtr.reference, r.reference_pubmed_id__doi_or_text)
                    if hasattr(r, 'literature_citation'):
                        lcc = bind_lcc(r.literature_citation())
                        asdf(s, ilxtr.literatureCitation, r.literature_citation, split=',', rdf_type=lcc)
                    if hasattr(r, 'expert_consultant'):
                        lcc = bind_lcc(r.expert_consultant())
                        asdf(s, ilxtr.expertConsultant, r.expert_consultant, split=',', rdf_type=lcc)

                    if hasattr(r, 'neuron_population_label_a_to_b_via_c'):
                        asdf(s, ilxtr.origLabel, r.neuron_population_label_a_to_b_via_c)
                        if csin is None:
                            asdf(s, skos.prefLabel, r.neuron_population_label_a_to_b_via_c)

                    if hasattr(r, 'subject') and r.subject().value.strip():
                        asdf(s, rdfs.label, r.subject, as_kwarg='label')
                        extra_kwargs[s]['override'] = True
                    elif working_set is not None:
                        asdf(s, rdfs.label, wrap(f'neuron type {_prefix} {_id}'), as_kwarg='label')
                        extra_kwargs[s]['override'] = True

                    if hasattr(r, 'alert_explanation'):
                        asdf(s, ilxtr.alertNote, r.alert_explanation)
                    if hasattr(r, 'alert_note'):
                        asdf(s, ilxtr.alertNote, r.alert_note)

                    if hasattr(r, 'explicit_complement') and r.explicit_complement().value:
                        p = map_predicates(r.relationship().value)
                        o = OntId(r.explicit_complement().value)
                        ec[(s, p)] = o

                    if hasattr(r, 'object_text') and r.object_text().value.strip():
                        p = (OntId(r.predicate_uri().value) if hasattr(r, 'predicate_uri') else
                             (r.predicate() if hasattr(r, 'predicate') else r.relationship()))
                        if p == ilxtr.expertConsultant:
                            asdf(s, p, r.object_text, rdf_type=rdflib.URIRef, split=';')
                        elif p == rdfs.label:
                            asdf(s, rdfs.label, r.object_text, as_kwarg='label')
                            extra_kwargs[s]['override'] = True
                        else:
                            asdf(s, p, r.object_text)

                    yes_cfu = hasattr(r, 'connected_from_uri') and r.connected_from_uri().value
                    yes_acp = hasattr(r, 'axonal_course_poset') and r.axonal_course_poset().value
                    if yes_cfu or yes_acp:
                        # s.u and OntId(...).u to avoid duplicate subjects/objects in the graph
                        # due to type vs instance issues for rdflib.URIRef and OntId
                        _v = r.object_uri().value if hasattr(r, 'object_uri') else r.identifier().value
                        if not _v:
                            _structure = r.structure().value
                            _alt_v = TEMP['MISSING_' + _structure.replace(' ', '-')]
                        else:
                            _structure = r.structure().value if hasattr(r, 'structure') else r.object().value
                            _alt_v = TEMP.BROKEN_EMPTY

                        if not debug and not _v:
                            raise ValueError(f'row missing object for {_structure}')

                        if _v and ',' in _v:
                            _r, _l = [_.strip() for _ in _v.split(',')]
                            #_v = _r  # FIXME TODO handle rl pairs
                            _obj = orders.rl(region=OntId(_r).u, layer=OntId(_l).u)
                        else:
                            try:
                                _obj = OntId(_v).u if _v else _alt_v
                            except OntId.UnknownPrefixError as e:
                                if debug:
                                    log.exception(e)
                                    _obj = TEMP.BROKEN_MALFORMED
                                else:
                                    raise e

                            _obj = orders.rl(region=_obj)

                        if yes_cfu:
                            for _frm in r.connected_from_uri().value.split(';'):
                                if ',' in _frm:
                                    _r, _l = [OntId(_.strip()).u for _ in _frm.split(',')]
                                    frm = orders.rl(region=_r, layer=_l)
                                else:
                                    frm = orders.rl(region=OntId(_frm.strip()).u)

                                cfdd[s.u].append((frm, _obj))

                        if yes_acp:
                            dd[s.u].append((int(r.axonal_course_poset().value), _obj))

                    if working_set:
                        class v:
                            value = working_set
                        asdf(s, ilxtr.inNLPWorkingSet, lambda x=v: x)

            except Exception as e:
                msg = f'\nbad row: | {" | ".join(r.values)} |'
                log.error(msg)
                raise e

    if cfdd:
        no_self = True
        if no_self:
            def _sigh(v):
                out = [ab for ab in v if ab[0] != ab[1]]
                if not out:
                    # interneurons basically
                    return v
                else:
                    return out

            sadj = {k:sorted(_sigh(v)) for k, v in cfdd.items()}
        else:
            sadj = {k:sorted(v) for k, v in cfdd.items()}
    elif dd:
        sorders = {k:sorted(v) for k, v in dd.items()}
        sadj = {k:ind_to_adj(v, neuron_id=k) for k, v in sorders.items()}
    else:
        sadj = {}

    snst = {k:orders.adj_to_nst(v) for k, v in sadj.items()}

    # XXX config must be called before creating any phenotypes
    # otherwise in_graph will not match when we go to serialize
    if config is None:
        config = Config('sparc-nlp')

    dd = defaultdict(list)
    for c, us, _s, _p, _o in trips:
        _, nlpns, working_set = snames[c.sheet_name]
        if isinstance(_o, LogicalPhenotype):
            olp = True
            real_o = _o
            _o = ''
        else:
            olp = False
            real_o = None

        for x in (_s, _p, _o):
            if re.match(r'(^[\s]+[^\s].*|.*[^\s][\s]+$)', x):
                msg = f'leading/trailing whitespace in {c} {_s!r} {_p!r} {_o!r}'
                log.error(msg)
                raise ValueError(msg)

        do_entail = 'neurdf/pred/ent/' in _p

        s = _s if isinstance(_s, OntId) else OntId(nlpns[_s])
        #log.debug(s)
        if isinstance(_p, OntId):
            p = map_uris(_p.u)
        else:
            try:
                p = map_predicates(_p)
            except KeyError as e:
                log.error(f'sigh {s}')
                raise e


        if not _o and not olp:  # handle empty cell case
            msg = f'missing object? {s.curie} {OntId(p).curie} ???'
            log.error(msg)
            if not debug:
                raise ValueError(msg)
            else:
                _o = TEMP.BROKEN_EMPTY

        elif nlpns is None:
            if ',' in _o:
                _r, _l = [_.strip() for _ in _o.split(',')]
                _o = IntersectionOf(OntId(_r).u, OntId(_l).u)
                #_o = _r  # FIXME TODO handle region/layer
        elif p == ilxtr.hasForwardConnectionPhenotype:
            if _o.startswith('http'):
                pass
            elif not isinstance(_o, OntId):
                _o = nlpns[_o]

        if p.endswith('hasComposerURI'):
            _ksn = _o.rsplit('/', 1)[-1]
            _o = f'https://composer.scicrunch.io/statement/{_ksn}'

        if olp:
            # TODO check pheno match
            dd[s, us].append(real_o)
            continue
        elif isinstance(_o, IntersectionOf):
            o = _o
        else:
            try:
                o = OntId(_o)
            except OntId.UnknownPrefixError as e:
                if debug:
                    log.exception(e)
                    o = TEMP.BROKEN_MALFORMED_2
                else:
                    raise e

        if p == owl.equivalentClass:
            to_add_append((s.u, p, o.u))
            continue

        if p in annotation_properties:
            to_add_append((s.u, p, o.u))
            continue

        try:
            if (do_entail or
                p == ilxtr.hasInstanceInTaxon or
                (p == ilxtr.hasAnatomicalSystemPhenotype and not ('aacar-7' in s.u or 'aacar-8' in s.u or s.u.endswith('keast-2') or s.u.endswith('keast-1'))) or  # XXX hardcoded workaround
                p == ilxtr.hasCircuitRolePhenotype or
                (p == ilxtr.hasFunctionalCircuitRolePhenotype and not ('sstom-11' in s.u or 'sstom-12' in s.u or 'sstom-13' in s.u or 'sstom-14' in s.u)) or  # XXX hardcoded workaround
                (p == ilxtr.hasProjectionPhenotype and not ('sdcol-l' in s.u or 'sdcol-j' in s.u))  # XXX hardcoded workaround for composer
                ):
                # XXX ilxtr.hasBiologicalSex is another candidate for this treatment
                # FIXME HACK, in order to be defining
                # we need something stronger than hasInstanceInTaxon for necessary and sufficient
                phen_class = EntailedPhenotype
            else:
                phen_class = Phenotype

            dd[s, us].append(phen_class(o, p))
        except TypeError as e:
            log.error(f'bad data for {c} {s} {p} {o}')
            raise e

        if ec and (s, p) in ec:
            dd[s, us].append(NegPhenotype(ec[(s, p)], p))


    # condense union sets
    dd2 = defaultdict(list)
    for (id, us), phenos in dd.items():
        if us is not None and us and len(phenos) > 1:
            phenos = LogicalPhenotype(OR, *phenos),

        dd2[id].extend(phenos)

    if neuron_class is None:
        neuron_class = NeuronSparcNlp

    sigh = []
    nrns = []
    ed = {}
    for id, phenos in dd2.items():
        nc = neuron_class if neuron_class_fun is None else neuron_class_fun(id)
        kwargs = extra_kwargs[id] if id in extra_kwargs else ed
        n = nc(*phenos, id_=id, **kwargs)
        if False and eff(n):
            n._sigh()  # XXXXXXXXXXXXXXXXXXXXXXXXXXXXXXX FIXME figure out why this is not getting called internally
            sigh.append(n)

        nrns.append(n)

    def sigh_bind(n, p):  # XXX FIXME UGH
        Phenotype.in_graph.bind(n, p)

    # FIXME these must also be in nifstd/scigraph/curie_map.yaml to show up on export
    sigh_bind('mmset4', snames['MM Set 4'][1])
    sigh_bind('semves', snames['Seminal Vesicles'][1])
    sigh_bind('prostate', snames['Prostate'][1])
    sigh_bind('femrep', snames['1. Female Reproductive-HUMAN'][1])
    sigh_bind('senmot', snames['All SM connections'][1])
    sigh_bind('kidney', snames['All KIDNEY connections'][1])
    sigh_bind('liver', snames['ALL Liver_Human_Rat_Mouse'][1])
    sigh_bind('swglnd', snames['Sheet1'][1])
    sigh_bind('comprt', nlp_ns('composer'))

    config.write()
    # skos no longer in rdflib default so not restored after cull_prefixes
    config._written_graph.namespace_manager.bind('skos', str(skos))
    labels = (
        #ilxtr.genLabel,
        #rdfs.label,
        ilxtr.localLabel, ilxtr.simpleLabel,
        ilxtr.simpleLocalLabel, skos.prefLabel)
    to_remove = [t for t in config._written_graph if t[1] in labels]
    [config._written_graph.remove(t) for t in to_remove]
    [config._written_graph.add(t) for t in to_add]
    add_partial_orders(config._written_graph, snst)
    for _n in nrns:
        if _n.id_ not in snst:
            msg = f'missing partial order for {_n.id_}'
            log.error(msg)
            continue

        _n.partialOrder(nested=snst[_n.id_])
    # uncomment to debug type vs instance issues
    #sigh = sorted(set(s for s in config._written_graph.subjects() if isinstance(s, rdflib.URIRef)))
    config._written_graph.write()
    config.write_python()
    return config,


if __name__ == '__main__':
    main()



import pytest
import importlib
from pyontutils.core import OntConjunctiveGraph
from .common import skipif_no_net, log, testing_base, _TestNeuronsBase


class TestNeurdf(_TestNeuronsBase):

    @pytest.mark.skipif(not importlib.find_loader('neurondm.compiled'), reason='no compiled sources')
    def test_models(self, debug=False):
        # XXX FIXME this test requires compiled to already have been populated
        # e.g. by running python -m neurondm.build models
        from neurondm.models import composer  # ensure composer curies are loaded
        from neurondm.core import uPREFIXES
        from neurondm.build import make_models
        configs = make_models(source='compiled')
        if debug:
            ncs = [(n, c) for c in configs for n in c.neurons()]
            ng = OntConjunctiveGraph()
            ng.namespace_manager.populate_from(uPREFIXES)
            done = set()
            for n, c in ncs:
                if n in done:
                    msg = f'duplicate neuron in config {c.name} {n.id_}'
                    log.debug(msg)
                    continue
                else:
                    done.add(n)

                for t in n._instance_neurdf():
                    try:
                        ng.add(t)
                    except Exception as e:
                        msg = f'neurdf error for {n}\n{t}'
                        log.error(msg)
                        log.exception(e)

            #ng.debug()
            path_debug = testing_base / 'neurdf-test-debug.ttl'
            ng.write(path_debug, format='nifttl')
            #breakpoint()

        # XXX BEWARE: rdf lists can appear twice in the serialization
        # if a neuron with the same id is present in more than one config
        g = OntConjunctiveGraph()
        g.namespace_manager.populate_from(uPREFIXES)
        for c in configs:
            c.neurdf_graph(graph=g)

        #g.debug()
        path = testing_base / 'neurdf-test.ttl'
        g.write(path, format='nifttl')
        #breakpoint()



import re
from collections import defaultdict
from urllib.parse import quote as url_quote
import rdflib
from pyontutils.sheets import Sheet
from pyontutils.namespaces import ilxtr, TEMP, rdfs, skos, owl, interlex_namespace
from neurondm.core import Config, NeuronEBM, Phenotype, EntailedPhenotype, NegPhenotype, log, OntCuries, OntId, add_partial_orders, IntersectionOf
from neurondm import orders

from neurondm.models.nlp import map_predicates, main as nlp_main


class NeuronBhuiyan2024(NeuronEBM):
    owlClass = ilxtr.NeuronBhuiyan2024
    shortname = 'Bhuiyan2024'
    _no_origLabel_hack = True


from sparcur.sheets import Sheet
class PRE(Sheet):
    _do_cache = True
    _re_cache = True
    name = 'npo-precision'
    sheet_name = 'NPO-Template (Individual cell types/species)'

from pyontutils.sheets import Row

def _subject_uri(self, nocite=True):
    gp = 'guinea pig'
    nid = self.neuron_id()
    if nocite and self.predicate_uri().value == 'ilxtr:literatureCitation':
        nid.value = ''
    else:
        v = nid.value
        if v and not v.startswith('http://'):
            pref = 'http://uri.interlex.org/tgbugs/uris/readable/neurons/bhuiyan/'
            if gp in v:
                v = v.replace(gp, 'guinea-pig')
            nid.value =  pref + v.lower().replace(' ', '/')

    return nid

def _proposed_action(self):
    return self.neuron_id()

def _predicate_uri(self):
    #return self.relation_type()
    return self.npo_property()

def _object_uri(self):
    return self.npo_property_value_iri()

def _object(self):
    if self.object_uri().value.strip():
        return self.property_value_label()
    else:
        return type('temp_cell', tuple(), {'value': ''})()

def _object_text(self):
    if not self.object_uri().value.strip():
        return self.property_value_label()
    else:
        return type('temp_cell', tuple(), {'value': ''})()

def _union_set(self):
    return self.union_set_number()

def ncbigene(nrns):
    pref = 'http://www.ncbi.nlm.nih.gov/gene/'
    c = nrns[0].out_graph.namespace_manager.curie
    ncbigene_ids = [c(p.p) for n in nrns for p in n.pes if p.p.startswith(pref)]
    print('\n'.join(ncbigene_ids))


def main():
    attrs = (
        'subject_uri',
        'proposed_action',
        'predicate_uri',
        'object_uri',
        'object_text',
        'object',
        'union_set',
    )
    saved = {a: getattr(Row, a) for a in attrs if hasattr(Row, a)}
    Row.subject_uri = _subject_uri
    Row.proposed_action = _proposed_action
    Row.predicate_uri = _predicate_uri
    Row.object_uri = _object_uri
    Row.object_text = _object_text
    Row.object = _object
    Row.union_set = _union_set

    config = Config('bhuiyan-2024')
    NeuronEBM.out_graph.add(
        (NeuronBhuiyan2024.owlClass,
         ilxtr.modelSource,
         OntId('https://doi.org/10.1126/sciadv.adj9173').u))

    cs = [PRE()]
    nlp_main(cs=cs, config=config, neuron_class=NeuronBhuiyan2024)
    nrns = config.neurons()
    #ncbigene(nrns)

    for a in attrs:  # restore old row state
        if a in saved:
            setattr(Row, saved[a])
        else:
            delattr(Row, a)

    return config,


if __name__ == '__main__':
    main()



from pyontutils import identity_bnode as idbn
from collections import Counter, defaultdict


def nst_to_adj(l):
    if not l:
        return []

    start = l[0]
    if (isinstance(start, list) or isinstance(start, tuple)):
        raise ValueError('malformed')

    first = [] if start == "blank" else [(start, e[0]) for e in l[1:] if e]
    second = [p for r in l[1:] if r for p in nst_to_adj(r)]
    return tuple(first + second)


def skey(ns):
    return (isinstance(ns, rl) and ns.layer is not None), ns


def assoc(k, ass):
    for a, b in ass:
        if a == k:
            return a, b


def dvals(k, ass):
    pair = assoc(k, ass)
    mem = ass[ass.index(pair):] if pair in ass else False
    tail = mem[1:] if mem else tuple()
    yield from ((pair[1], *dvals(k, tail)) if tail else pair[1:]) if pair else tuple()


def inest(adj, start, seen, expect, starts):
    nl = tuple()
    out = []
    maybe_out = []
    seen[start] += 1
    for next_start in sorted(dvals(start, adj), key=skey):
        if next_start in seen and seen[next_start] > 0:
            # do not append if next_start already seen more than expected
            nl, snext = (next_start,), seen
            if seen[next_start] < expect[next_start]:
                out.append(nl)
                seen[next_start] += 1
            elif next_start in starts and seen[next_start] == expect[next_start]:
                out.append(nl)
            else:
                pass  # pretty sure should never get here now
        else:
            nl, snext = inest(adj, next_start, seen, expect, starts)
            out.append(nl)

        seen = snext

    l = tuple(out)
    return ((start, *l) if l else (start,)), seen


def adj_to_nst(adj, start=None):
    if not adj:
        return tuple()

    keys = set([a[0] for a in adj])
    values = set([
        a[1] for a in adj
        # filter reflexive cases so they can be correctly ided as starts
        # if we have soma, dendrite, and axon all located in the same region
        # then it looks like we have an a -> a case for the starting point
        # therefore we ignore cycles of length zero when searching for starts
        if a[0] != a[1]
    ])

    #inverted = [(a[1], a[0]) for a in adj]  # unused
    starts = keys - values
    #ends = values - keys  # unused

    if not starts:
        # the graph forms a full cycle with all potential starting points
        # involved in a cycle, therefore pick an arbitrary but stable start
        # XXX WARNING this fails if there is a cycle length zero at a true start
        starts = {sorted(adj)[0][0]}

    seen = {thing: 0 for thing in (keys | values)}
    expect = dict(Counter([a[1] for a in adj]))
    nl = None
    out = []
    for start in sorted(starts, key=skey):
        nl, snext = inest(adj, start, seen, expect, starts)
        seen = snext
        out.append(nl)

    l = tuple(out)
    if len(l) == 1 and (isinstance(l[0], list) or isinstance(l[0], tuple)):
        return l[0]
    else:
        return ["blank", *l]


def lin_to_adj(distinct_paths, linkers=tuple()):
    adj = []
    for linear in distinct_paths:
        adj.extend(tuple(zip(linear[:-1], linear[1:])))

    adj.extend(linkers)
    # [(t, c) for t, c in sigh.most_common() if c > 1]
    # FIXME the cycle detection in tsort in adj_to_lin is not working ???
    if len(adj) != len(set(adj)):
        sigh = Counter(adj)
        msg = f'duplicate paths in linear representation\n{adj}'
        raise Exception(msg)

    return tuple(adj)


def adj_to_lin(adj):
    # this is not deterministic when there are runs of equal length
    # this should return a list of linear paths from longest to
    # shortest that are sufficient to cover all nodes, essentially
    # the longest single path through the tree and then branches

    # yay dags have a linear time algo
    tsorted = idbn.toposort(adj, unmarked_key=(lambda t: (t if isinstance(t, rl) else rl(t, None))))

    _dd = defaultdict(list)
    [_dd[b].append(a) for a, b in adj]
    incoming = dict(_dd)

    if len(tsorted) != len(set(tsorted)):
        raise Exception('sigh al')

    paths = {n: [n] for n in tsorted}
    for v in tsorted:
        if v in incoming:
            for w in incoming[v]:
                lpw, lpv = len(paths[w]), len(paths[v])
                if lpw <= lpv:
                    # always add the w to paths[v] never extend
                    # paths[w] it works in a subset of cases but is
                    # fundamentally incorrect and can lead to cycles
                    paths[w] = [w] + paths[v]

    seen = set()
    distinct = set()
    linkers = set()  # extra linkers are only needed if there is more than one incoming edge for short paths
    linked = set()
    spaths = sorted([tuple(v) for v in paths.values()], key=lambda v: len(v), reverse=True)
    _intermediate = set()
    for path in spaths:
        start = path[0]
        end = path[-1]
        if start not in seen and end not in seen:
            if start in incoming:
                inc = incoming[start]
                _path = inc[0], *path
                distinct.add(_path)
                _intermediate.update(list(zip(_path[:-1], _path[1:])))
                linkers.update([(i, start) for i in inc[1:]])
                linked.add(start)
            else:
                distinct.add(path)
                # FIXME watch out for going quadratic because of this
                _intermediate.update(list(zip(path[:-1], path[1:])))

            seen.update(path)

        elif start not in seen and end in seen:  # multi-parent case
            if start in incoming:
                inc = incoming[start]
                path_prefix = [inc[0]] if inc[0] not in seen else []
            else:
                inc = None
                path_prefix = []
            for n in path:
                if n in seen:
                    linkers.add((_prev, n))
                    # put the break before append because
                    # trying to detect overlap is an expensive operation
                    # and this way is consistent with the multi-child case
                    break

                _prev = n
                path_prefix.append(n)

            distinct.add(tuple(path_prefix))
            _intermediate.update(list(zip(path_prefix[:-1], path_prefix[1:])))
            if inc is not None:
                linkers.update([(i, start) for i in inc])
                linked.add(start)
            seen.update(path_prefix)

        elif start in incoming and start not in linked:  # multiparent case
            # edge case where you have ((1 2) (1 3) (4 2) (4 5)) and (4 2) is not included
            # but all other potential linkers are so you have to skip potential linkers that
            # are already in the distinct ... the other possibility would be to simply include
            # the linkers in distinct, except I think it is clearer to keep them separate?
            # consider the design tradeoff
            linkers.update(
                [(i, start) for i in incoming[start]
                 # FIXME surely there is a better way to do this than to check intermediates
                 if (i, start) not in distinct and (i, start) not in _intermediate])
            linked.add(start)

    return distinct, linkers


def bind_rdflib():
    import rdflib
    from pyontutils.namespaces import rdf

    def to_rdf(g, nested):
        if isinstance(nested, list) or isinstance(nested, tuple):
            bn0 = rdflib.BNode()
            rdflib.collection.Collection(g, bn0, (to_rdf(g, n) for n in nested))
            return bn0
        elif isinstance(nested, rl):
            bn1 = rdflib.BNode()
            return nested.to_rdf(to_rdf, g, bn1)
        elif isinstance(nested, rdflib.term.Node):
            return nested
        else:
            return rdflib.Literal(nested)

    def from_rdf(g, linker, to_python=False):
        nested = {}
        for s, o in g[:linker:]:
            l = tuple(getlist(g, o, to_python=to_python))
            nested[s] = l

        return nested

    def getlist(g, bn0, to_python=False):
        def f(node, g):
            o = None

            #for o in g[node:rdf.first:]:
            for o in g.objects(node, rdf.first):
                if isinstance(o, rdflib.BNode):
                    # unfortunately we have to branch on the result type
                    # or have to look ahead to see if the list continues
                    # picked the look ahead so that we don't have to check
                    # length and we don't have to know about the rl type
                    if list(g[o:rdf.first]):  # list continues
                        # XXX watch out for malformed lists that might not
                        # have a rdf:first ?
                        yield tuple(getlist(g, o, to_python=to_python))
                    else:  # likely [ :a :b ] case
                        yield from getlist(g, o, to_python=to_python)

                else:
                    if to_python and isinstance(o, rdflib.Literal):
                        yield o.toPython()
                    else:
                        yield o

            #for o in g[node:rdf.rest:]:
            for o in g.objects(node, rdf.rest):
                if o != rdf.nil:
                    yield from getlist(g, o, to_python=to_python)

            if o is None and isinstance(node, rdflib.BNode):  # [ :a  :b ] case
                #for p, o in g[node:]:
                for p, o in g.predicate_objects(node):
                    if to_python:
                        yield rl((p.toPython() if isinstance(p, rdflib.Literal) else p),
                                (o.toPython() if isinstance(o, rdflib.Literal) else o),)
                    else:
                        yield rl(p, o)

        yield from g.transitiveClosure(f, bn0)

    return to_rdf, from_rdf


class rl:
    def __init__(self, region, layer=None):
        self.region = region
        self.layer = layer

    def __repr__(self):
        if self.layer:
            return f"{self.__class__.__name__}({self.region!r}, {self.layer!r})"
        else:
            return f"{self.__class__.__name__}({self.region!r})"

    def to_rdf(self, to_rdf, g, bn):
        r = to_rdf(g, self.region)
        if self.layer is None:
            return r

        l = to_rdf(g, self.layer)
        g.add((bn, r, l))
        return bn

    def __eq__(self, other):
        return type(self) == type(other) and self.region == other.region and self.layer == other.layer

    def __hash__(self):
        return hash((self.__class__, self.region, self.layer))

    def __lt__(self, other):
        if type(self) == type(other):
            sreg = '' if self.region is None else self.region
            slay = '' if self.layer is None else self.layer
            oreg = '' if other.region is None else other.region
            olay = '' if other.layer is None else other.layer
            return sreg < oreg or slay < olay

    def hash_thing(self):
        pair = self.region, self.layer
        i = idbn.IdentityBNode(pair)
        prefix_len = 10
        a = i.cypher.__name__.split('_')[-1]
        b = i.identity.hex()[:prefix_len]
        return f'{a}_{prefix_len}:{b}'


def test():
    from pprint import pprint
    to_rdf, from_rdf = bind_rdflib()
    adj_test = (
        ("a", "b"),
        ("b", "c"),
        ("c", "d"),
        ("d", "e"),
        ("d", "k"),

        ("a", "f"),
        ("f", "g"),

        ("h", "i"),
        ("i", "j"),
    )

    nested = adj_to_nst(adj_test)
    linlin = adj_to_lin(adj_test)
    print('nested', nested)
    print('asdf', list(dvals('a', adj_test)))
    print('asdf', list(dvals('c', adj_test)))
    print('asdf', list(dvals('d', adj_test)))
    print('linlin', linlin)

    adj_out = nst_to_adj(nested)
    print(adj_out)

    assert set(adj_test) == set(adj_out), 'oops'

    from pyontutils.core import OntGraph
    from pyontutils.namespaces import ilxtr, rdf

    g = OntGraph()


    bn = to_rdf(g, nested)
    g.add((ilxtr['sub-1'], ilxtr.predicate, bn))

    adj_2 = (
        (ilxtr.a, ilxtr.b),
        (ilxtr.b, ilxtr.c),
        (ilxtr.c, ilxtr.d),
        (ilxtr.d, ilxtr.e),
        (ilxtr.d, ilxtr.h),

        (ilxtr.a, ilxtr.f),
        (ilxtr.f, ilxtr.g),
    )
    nst_2 = adj_to_nst(adj_2)
    lln_2 = adj_to_lin(adj_2)
    print('lln_2', lln_2)
    print('nst_2', nst_2)
    bn = to_rdf(g, nst_2)
    g.add((ilxtr['sub-2'], ilxtr.predicate, bn))

    lin_3 = ([1, 2, 3, 4, 5, 6],)
    adj_3 = lin_to_adj(lin_3)
    nst_3 = adj_to_nst(adj_3)
    lln_3 = adj_to_lin(adj_3)
    assert lin_3 == tuple(list(l) for l in lln_3[0]) and not lln_3[1], (lin_3, tuple(list(l) for l in lln_3[0]), lln_3[1])
    print('nst_3', nst_3)
    print('lln_3', lln_3)
    bn = to_rdf(g, nst_3)
    g.add((ilxtr['sub-3'], ilxtr.predicate, bn))

    lin_4 = ([1, 2, 3, 4, 5], [3, 6, 7],)
    adj_4 = lin_to_adj(lin_4)
    nst_4 = adj_to_nst(adj_4)
    lln_4 = adj_to_lin(adj_4)
    assert lin_4 == tuple(list(l) for l in lln_4[0]) and not lln_4[1], (lin_4, tuple(list(l) for l in lln_4[0]), lln_4[1])
    print('lln_4', lln_4)
    print('nst_4', nst_4)
    bn = to_rdf(g, nst_4)
    g.add((ilxtr['sub-4'], ilxtr.predicate, bn))

    adj_5 = (
        (rl(ilxtr.a,), rl(ilxtr.b,)),
        (rl(ilxtr.b,), rl(ilxtr.c, ilxtr.l1)),
        #(rl(ilxtr.c, ilxtr.l1), rl(ilxtr.d,)),  # alt end # this one does not duplicate the final node in adj_to_nst
        (rl(ilxtr.c, ilxtr.l1), rl(ilxtr.c, ilxtr.l2)),
        (rl(ilxtr.c, ilxtr.l2), rl(ilxtr.c, ilxtr.l3)),
        (rl(ilxtr.c, ilxtr.l3), rl(ilxtr.d,)),
    )
    nst_5 = adj_to_nst(adj_5)
    lln_5 = adj_to_lin(adj_5)
    pprint(('lln_5', lln_5))
    pprint(('nst_5', nst_5))
    bn = to_rdf(g, nst_5)
    g.add((ilxtr['sub-5'], ilxtr.predicate, bn))

    # multi-parent multi-child case
    adj_6 = (
        (3, 1),
        (3, 2),
        (4, 1),
        (4, 2),
        (5, 1),
        (5, 2),
        (6, 1),
        (6, 2),
        (1, 7),
        (1, 8),
        (2, 7),
        (2, 8),
    )
    nst_6 = adj_to_nst(adj_6)
    lln_6 = adj_to_lin(adj_6)
    rej_6 = nst_to_adj(nst_6)
    rel_6 = lin_to_adj(*lln_6)
    pprint(('nst_6', nst_6))
    pprint(('adj_6', adj_6))
    pprint(('rej_6', rej_6))
    pprint(('rel_6', rel_6))
    pprint(('lln_6', lln_6))
    assert sorted(adj_6) == sorted(rej_6) == sorted(rel_6), breakpoint()
    bn = to_rdf(g, nst_6)
    g.add((ilxtr['sub-6'], ilxtr.predicate, bn))

    un_rdf = from_rdf(g, ilxtr.predicate, to_python=True)
    assert un_rdf[ilxtr['sub-1']] == tuple(nested)
    assert un_rdf[ilxtr['sub-2']] == tuple(nst_2)
    assert un_rdf[ilxtr['sub-3']] == tuple(nst_3)
    assert un_rdf[ilxtr['sub-4']] == tuple(nst_4)
    un_rdf[ilxtr['sub-5']] == tuple(nst_5)  # non-invertable case due to layer=None ambiguity
    assert un_rdf[ilxtr['sub-6']] == tuple(nst_6)

    adj_raw_7 = [  # aacar-6 (confusingly)
        # the issue here is that the graph is not fully connected
        # AND that some of the disconnected parts have auto cycles
        (('ILX:0774266', None), ('ILX:0787082', None)),
        (('ILX:0786141', None), ('ILX:0784439', None)),
        (('ILX:0786228', None), ('ILX:0791105', None)),
        (('ILX:0786272', None), ('ILX:0788945', None)),
        (('ILX:0786722', None), ('ILX:0787562', None)),
        (('ILX:0789947', None), ('ILX:0787946', None)),
        (('ILX:0793555', None), ('UBERON:0000948', 'UBERON:0015129')),
        (('ILX:0793556', None), ('UBERON:0000948', 'UBERON:0015129')),
        (('UBERON:0000948', 'UBERON:0002348'), ('UBERON:0000948', 'UBERON:0002348')),
        (('UBERON:0000948', 'UBERON:0002348'), ('UBERON:0002078', 'UBERON:0002349')),
        (('UBERON:0000948', 'UBERON:0002348'), ('UBERON:0002079', 'UBERON:0002349')),
        (('UBERON:0000948', 'UBERON:0002348'), ('UBERON:0002080', 'UBERON:0002349')),
        (('UBERON:0000948', 'UBERON:0002348'), ('UBERON:0002084', 'UBERON:0002349')),
        (('UBERON:0000948', 'UBERON:0015129'), ('ILX:0793555', None)),
        (('UBERON:0000948', 'UBERON:0015129'), ('ILX:0793556', None)),
        (('UBERON:0000948', 'UBERON:0015129'), ('UBERON:0000948', 'UBERON:0002348')),
        (('UBERON:0002078', 'UBERON:0002349'), ('UBERON:0002078', 'UBERON:0002165')),
        (('UBERON:0002079', 'UBERON:0002349'), ('UBERON:0002079', 'UBERON:0002165')),
        (('UBERON:0002080', 'UBERON:0002349'), ('UBERON:0002080', 'UBERON:0002165')),
        (('UBERON:0002084', 'UBERON:0002349'), ('UBERON:0002084', 'UBERON:0002165')),
    ]

    adj_raw_7 = [
        # the issue here is the auto cyclic nodes
        # when all these are run reg_7 -> empty
        #(('ILX:0793556', None), ('ILX:0793556', None)),  # cycle here
        # somehow the autocycle will cause output issues if it is placed in the 0th position in the list
        # but not down below ??? answer: because rl.__lt__ was using and instead of or (derp)
        (('UBERON:0000948', 'UBERON:0002348'), ('UBERON:0000948', 'UBERON:0002348')),  # cycle here
        (('ILX:0793556', None), ('UBERON:0000948', 'UBERON:0015129')),
          (('UBERON:0000948', 'UBERON:0015129'), ('UBERON:0000948', 'UBERON:0002348')),  # when only this and cycle it works ish ???
            #(('UBERON:0000948', 'UBERON:0002348'), ('UBERON:0000948', 'UBERON:0002348')),  # cycle here
            (('UBERON:0000948', 'UBERON:0002348'), ('UBERON:0002084', 'UBERON:0002349')),
              (('UBERON:0002084', 'UBERON:0002349'), ('UBERON:0002084', 'UBERON:0002165')),
            (('UBERON:0000948', 'UBERON:0002348'), ('UBERON:0002080', 'UBERON:0002349')),
              (('UBERON:0002080', 'UBERON:0002349'), ('UBERON:0002080', 'UBERON:0002165')),
    ]
    adj_7 = tuple(tuple(rl(*(a if a is None else a for a in _rl)) for _rl in rls) for rls in adj_raw_7)
    nst_7 = adj_to_nst(adj_7)
    #lln_7 = adj_to_lin(adj_7)  # cycle
    rej_7 = nst_to_adj(nst_7)
    pprint(('nst_7', nst_7))
    pprint(('adj_7', sorted(adj_7)))
    pprint(('rej_7', sorted(rej_7)))
    #pprint(('lln_7', lln_7))
    assert sorted(adj_7) == sorted(rej_7)
    bn = to_rdf(g, nst_7)
    g.add((ilxtr['sub-7'], ilxtr.predicate, bn))

    adj_raw_8 = [
        # this produces complete nonsense
        # XXX why the, how the, heck does this produce the result ... (answer: only check equality of region)
        (('UBERON:0006450', 'UBERON:0002318'), ('UBERON:0006450', 'UBERON:0002181')),
        (('UBERON:0006450', 'UBERON:0002318'), ('UBERON:0006450', 'UBERON:0004677')),
        (('UBERON:0006450', 'UBERON:0002318'), ('UBERON:0006450', 'UBERON:0006118')),
        (('UBERON:0006450', 'UBERON:0002318'), ('UBERON:0006450', 'UBERON:0016576')),
        (('UBERON:0006450', 'UBERON:0002318'), ('UBERON:0006450', 'UBERON:0016578')),
    ]
    adj_8 = tuple(tuple(rl(*(a if a is None else a for a in _rl)) for _rl in rls) for rls in adj_raw_8)
    adj_alt_8 = tuple(tuple('-'.join(('' if a is None else a for a in _rl)) for _rl in rls) for rls in adj_raw_8)
    nst_alt_8 = adj_to_nst(adj_alt_8)  # XXX the issue is clearly in rl
    nst_raw_8 = adj_to_nst(adj_raw_8)  # XXX the issue is clearly in rl
    he = hash(adj_8[0][0]) == hash(adj_8[1][0])
    ee = adj_8[0][0] == adj_8[1][0]
    #rej_raw_8 = nst_to_adj(nst_raw_8)  # though we can't convert tuples back because of type issues
    nst_8 = adj_to_nst(adj_8)
    lln_8 = adj_to_lin(adj_8)
    rej_8 = nst_to_adj(nst_8)
    rel_8 = lin_to_adj(*lln_8)
    pprint(('nst_8', nst_8), width=120)
    pprint(('adj_8', adj_8), width=120)
    pprint(('rej_8', rej_8), width=120)
    pprint(('lln_8', lln_8))
    assert sorted(adj_8) == sorted(rej_8) == sorted(rel_8)
    bn = to_rdf(g, nst_8)
    g.add((ilxtr['sub-8'], ilxtr.predicate, bn))

    adj_raw_9 = [  # from keast 11 branching issue 18683 same as for adj_raw_8
        (('UBERON:0002856', None), ('UBERON:0006450', None #'UBERON:0002318'
                                    )),
        (('UBERON:0002856', None), ('UBERON:0018683', None)),
        (('UBERON:0002857', None), ('UBERON:0006448', None #'UBERON:0002318'
                                    )),
        (('UBERON:0002857', None), ('UBERON:0018683', None)),

        #(('UBERON:0018683', None), ('UBERON:0005453', None)),
    ]
    adj_9 = tuple(tuple(rl(*(a if a is None else a for a in _rl)) for _rl in rls) for rls in adj_raw_9)
    nst_9 = adj_to_nst(adj_9)
    lln_9 = adj_to_lin(adj_9)
    rej_9 = nst_to_adj(nst_9)
    rel_9 = lin_to_adj(*lln_9)
    pprint(('nst_9', nst_9))
    pprint(('adj_9', adj_9))
    pprint(('rej_9', rej_9))
    pprint(('lln_9', lln_9))
    assert sorted(adj_9) == sorted(rej_9) == sorted(rel_9), [a for a in adj_9 if a not in rej_9]
    bn = to_rdf(g, nst_9)
    g.add((ilxtr['sub-9'], ilxtr.predicate, bn))

    adj_10 = [
        (1, 2),
        (2, 3),
        (3, 4),
        (4, 5),
        (5, 6),
        (5, 2),  # long cycle not including start
        (6, 1),  # long cycle including "start"
    ]
    nst_10 = adj_to_nst(adj_10)
    rej_10 = nst_to_adj(nst_10)
    pprint(('nst_10', nst_10))
    pprint(('adj_10', adj_10))
    pprint(('rej_10', rej_10))
    assert sorted(adj_10) == sorted(rej_10), [a for a in adj_10 if a not in rej_10]
    bn = to_rdf(g, nst_10)
    g.add((ilxtr['sub-10'], ilxtr.predicate, bn))

    lin_11 = [
        [0, 1, 2,       5, 6, 7, 8,],
        [      2, 3, 4, 5,         ],
    ]
    adj_11 = lin_to_adj(lin_11)
    lln_11 = adj_to_lin(adj_11)

    adj_12 = (
        (1, 3),
        (1, 2),
        (2, 4),)
    lln_12 = adj_to_lin(adj_12)
    rel_12 = lin_to_adj(*lln_12)
    assert sorted(adj_12) == sorted(rel_12), breakpoint()

    g.debug()

    # cycle tests
    adj_to_nst([(1, 2), (2, 3), (2, 2)])  # ok because does not include start
    adj_to_nst([(1, 2), (2, 3), (3, 3)])  # ok because does not include start
    adj_to_nst([(4, 2), (1, 2), (2, 3), (3, 1)])  # ok because does not include all starts
    adj_to_nst([(4, 2), (1, 2), (2, 3), (3, 1), (3, 4)])  # includes all start node
    adj_to_nst([(1, 2), (2, 3), (3, 1)])  # includes all start node
    adj_to_nst([(1, 2), (2, 3), (1, 1)])  # includes all start nodes

    # make sure reflexive nodes (cycle length zero) can still be used as a start
    adj_wat = [
        (rl('1'), rl('1')),
        (rl('0', '2'), rl('0', '3')),
        (rl('0', '2'), rl('1')),
        (rl('d', 's'), rl('d', 's')),
        (rl('d', 's'), rl('d', 's')),
        (rl('d', 's'), rl('d', '4')),
        (rl('d', '5'), rl('0', '2')),
        (rl('0', '3'), rl('0', '6')),
        (rl('O', 's'), rl('O', 's')),
        (rl('O', 's'), rl('O', 's')),
        (rl('O', 's'), rl('O', '4')),
        (rl('O', '5'), rl('0', '2')),
        (rl('d', '4'), rl('d', '5')),
        (rl('O', '4'), rl('O', '5')),
    ]
    nst_wat = adj_to_nst(adj_wat)
    adj_watn = nst_to_adj(nst_wat)
    assert rl('O', 's') in set(e for es in adj_watn for e in es), f'rt failed'
    snst_wat = adj_to_nst(sorted(adj_wat))
    adj_watsn = nst_to_adj(snst_wat)
    assert rl('O', 's') in set(e for es in adj_watn for e in es), f'rt failed'


if __name__ == '__main__':
    test()



import unittest
from pyontutils.core import OntGraph
from pyontutils.namespaces import ilxtr
#from neurondm import simple
from neurondm.core import (Neuron, Phenotype, IntersectionOf, IntersectionOfPartOf, UnionOf,
                           Config, LogicalPhenotype, OR, EntailedPhenotype, EntailedLogicalPhenotype)
# FIXME for offline testing
try:
    #config = Config('owl-object-test', import_no_net=True)
    pass
except:
    raise
    Phenotype._location_predicates = (
        ilxtr.hasAxonLocatedIn,
    )
    Phenotype.part_of_graph = OntGraph()


class TestOwlObject(unittest.TestCase):
    _Phenotype = Phenotype
    _LogicalPhenotype = LogicalPhenotype

    def setUp(self):
        self.Phenotype = self._Phenotype
        self.LogicalPhenotype = self._LogicalPhenotype
        self.config = Config('owl-object-test', import_no_net=True)

    def _do(self, n):
        print(n)
        print(self.config.python())
        print(self.config.ttl())

        # this testing call needs to happen after we test ttl
        # and python out due to the braindead stateful nature of
        # infixowl
        g = OntGraph()
        n._graphify(graph=g)
        g.debug()

    def test_uo(self):
        n = Neuron(self.Phenotype(UnionOf(ilxtr.somaloc1, ilxtr.somaloc2, ilxtr.somaloc3), ilxtr.hasSomaLocatedIn))
        self._do(n)

    def test_io(self):
        n = Neuron(self.Phenotype(IntersectionOf(ilxtr.thing1, ilxtr.thing2)))
        self._do(n)

    def test_io_loc(self):
        n = Neuron(self.Phenotype(IntersectionOf(ilxtr.region, ilxtr.layer), ilxtr.hasAxonLocatedIn))
        self._do(n)

    def test_iopo(self):
        # TODO should iopo check whether the predicate is a location predicate?
        n = Neuron(self.Phenotype(IntersectionOfPartOf(ilxtr.region, ilxtr.layer)))
        self._do(n)

    # mixed

    def test_uo_m(self):
        n = Neuron(self.Phenotype(UnionOf(ilxtr.somaloc1, ilxtr.somaloc2, ilxtr.somaloc3), ilxtr.hasSomaLocatedIn),
                   self.Phenotype(ilxtr.other))
        self._do(n)

    def test_io_m(self):
        n = Neuron(self.Phenotype(IntersectionOf(ilxtr.thing1, ilxtr.thing2)),
                   self.Phenotype(ilxtr.other))
        self._do(n)

    def test_io_loc_m(self):
        n = Neuron(self.Phenotype(IntersectionOf(ilxtr.region, ilxtr.layer), ilxtr.hasAxonLocatedIn),
                   self.Phenotype(ilxtr.other))
        self._do(n)

    def test_iopo_m(self):
        # TODO should iopo check whether the predicate is a location predicate?
        n = Neuron(self.Phenotype(IntersectionOfPartOf(ilxtr.region, ilxtr.layer)),
                   self.Phenotype(ilxtr.other))
        self._do(n)

    # mixed with logical

    def test_uo_m(self):
        n = Neuron(self.Phenotype(UnionOf(ilxtr.somaloc1, ilxtr.somaloc2, ilxtr.somaloc3), ilxtr.hasSomaLocatedIn),
                   self.LogicalPhenotype(OR,
                                         self.Phenotype(ilxtr.other),
                                         self.Phenotype(ilxtr.another)))
        self._do(n)

    def test_io_m(self):
        n = Neuron(self.Phenotype(IntersectionOf(ilxtr.thing1, ilxtr.thing2)),
                   self.LogicalPhenotype(OR,
                                         self.Phenotype(ilxtr.other),
                                         self.Phenotype(ilxtr.another)))
        self._do(n)

    def test_io_loc_m(self):
        n = Neuron(self.Phenotype(IntersectionOf(ilxtr.region, ilxtr.layer), ilxtr.hasAxonLocatedIn),
                   self.LogicalPhenotype(OR,
                                         self.Phenotype(ilxtr.other),
                                         self.Phenotype(ilxtr.another)))
        self._do(n)

    def test_iopo_m(self):
        # TODO should iopo check whether the predicate is a location predicate?
        n = Neuron(self.Phenotype(IntersectionOfPartOf(ilxtr.region, ilxtr.layer)),
                   self.LogicalPhenotype(OR,
                                         self.Phenotype(ilxtr.other),
                                         self.Phenotype(ilxtr.another)))
        self._do(n)


class TestEntailedOwlObject(TestOwlObject):
    _Phenotype = EntailedPhenotype
    _LogicalPhenotype = EntailedLogicalPhenotype


class TestNestOwlObject(unittest.TestCase):
    # so in theory (aka owl) it is possible to nest intersectionOf and unionOf arbitrariliy however
    # the practical use cases where that ability is needed are currently not known all the use
    # cases for OwlObject essentially end at a single leve I suppose there might be a case where we
    # have (uo (io region-1 layer-1) (io region-2 layer-2)) so we could implement it, noting that
    # other tools (like composer) currently only have support for the equivalent of (io region
    # layer) and support for (uo soma-loc-1 soma-loc-2 ...)  only implicitly when using the
    # hasSomaLocatedIn property

    # we also generally want to discourage the use of OwlObjects because the whole point of bagging
    # phenotypes is to keep the phenotypes as simple as possible

    # we need (iopo region layer) to achieve parity with ApiNATOMY semantics and to reduce the need
    # to mint identifiers for simple intersection cases, but union-of actually poses something of
    # an interpretational challeng for partial orders, even though it does clarify the fact that
    # there is only one soma location for any individual type, and is also technically necessary to
    # work around the fact that having multiple soma locations makes the owl classes unsatisfiable
    # for any individual real neuron because the top level restriction is an intersection of
    # multiple discrete anatomical locations, we could use LogicalPhenotype for somal location, but
    # OwlObject union of seems to make more sense? ... and thus the challenge

    _Phenotype = Phenotype
    _LogicalPhenotype = LogicalPhenotype

    setUp = TestOwlObject.setUp
    _do = TestOwlObject._do

    def test_uo_io(self):
        n = Neuron(self.Phenotype(UnionOf(IntersectionOf(ilxtr.a, ilxtr.b), IntersectionOf(ilxtr.c, ilxtr.d))))
        self._do(n)

    def test_uo_io_loc(self):
        n = Neuron(self.Phenotype(UnionOf(IntersectionOf(ilxtr.a, ilxtr.b),
                                          IntersectionOf(ilxtr.c, ilxtr.d)),
                                  ilxtr.hasAxonLocatedIn))
        self._do(n)


class TestEntailedNestOwlObject(TestNestOwlObject):
    _Phenotype = EntailedPhenotype
    _LogicalPhenotype = EntailedLogicalPhenotype



import csv
import rdflib
import augpathlib as aug
from pyontutils import sheets
from pyontutils.namespaces import TEMP, ilxtr, rdfs, skos, interlex_namespace
from neurondm.models.nlp import map_predicates, main as nlp_main, NeuronSparcNlp
from neurondm.models.apinat_npo import NeuronApinatSimple, NeuronApinatComplex
from neurondm.core import log as _log, uPREFIXES, Config, Neuron

log = _log.getChild('composer')
anat_space_hack = 'http://purl.obolibrary.org/obo/UBERON_0000464'
# run at top level to ensure these are registered
uPREFIXES['gastint'] = 'http://uri.interlex.org/composer/uris/set/gastint/'
uPREFIXES['pain1'] = 'http://uri.interlex.org/composer/uris/set/pain1/'
uPREFIXES['portal'] = 'http://uri.interlex.org/composer/uris/set/portal/'

ilxcr = rdflib.Namespace(interlex_namespace('composer/uris/readable/'))

def get_csv_sheet(path):
    with open(path, 'rt') as f:
        _rows = list(csv.reader(f))

    # remove rows that are missing a value since the export doesn't do
    # that and we will get missing if not provided
    # FIXME TODO take NLP-ID and promote to subject if missing
    idx = _rows[0].index('Object URI')
    oidx = _rows[0].index('Object')
    sidx = _rows[0].index('Subject URI')
    pidx = _rows[0].index('Predicate URI')
    tidx = _rows[0].index('Object Text')
    ridx = _rows[0].index('Reference (pubmed ID, DOI or text)')
    eidx = _rows[0].index('Expert Consultant')
    cidx = _rows[0].index('Connected from uri')
    state_idx = _rows[0].index('Statement State')
    derp_idx = _rows[0].index('Object')
    _rows[0][ridx] = 'literature citation'  # XXX HACK
    sigh = 'https://uri.interlex.org/'
    sigh2 = 'https://scicrunch.org/scicrunch/interlex/view/'
    bad = 'https://rrid.site/data/record/nlx_144509-1/RRID:SCR_018709/resolver?q=SAWG&i=rrid:scr_018709'
    bad2 = 'https://rrid.site/resolver/SCR_018709'
    expcon = 'http://uri.interlex.org/tgbugs/uris/readable/expertConsultant'
    doibad = 'http://dx.doi.org/'
    doibads = (
        ('https://analyticalsciencejournals.onlinelibrary.wiley.com/doi/10.1002/jemt.10225', 'https://doi.org/10.1002/jemt.10225'),
        ('https://onlinelibrary.wiley.com/doi/abs/10.1002/cphy.c140032', 'https://doi.org/10.1002/cphy.c140032'),
        ('https://physoc.onlinelibrary.wiley.com/doi/abs/10.1113/JP271869','https://doi.org/10.1113/JP271869'),
    )
    nogitl = 'https://github.com/open-physiology/apinatomy-models'
    aacar_1_5_issue = 'https://uri.interlex.org/composer/uris/set/aacar/5'
    for i, _r in enumerate(_rows[1:]):  # FIXME EVEN BIGGER HACK
        if (not (_r[idx] or _r[tidx])) and _r[derp_idx]:
            _r[tidx] = _r[derp_idx]

        if _r[cidx] == anat_space_hack:
            _r[cidx] = ''

        if _r[idx] == 'http://uri.interlex.org/tgbugs/uris/readable/HasProjection':
            # something got mangled inside composer ...
            _r[idx] = 'http://uri.interlex.org/tgbugs/uris/readable/ProjectionPhenotype'
            log.error(f'bad object value for {i + 2}')

        if _r[sidx] == aacar_1_5_issue:  # XXX fix for bad working set interaction
            _r[sidx] = 'http://uri.interlex.org/tgbugs/uris/readable/neuron-type-aacar-1'
            _r[0] = 'neuron type aacar 1'
            _r[state_idx] = 'wat'

        if _r[derp_idx] == 'not specified':
            _r[sidx] = ''  # for skip the row to avoid bad triples with TEMP:MISSING
            #_r[derp_idx] = ''

        if doibad in _r[ridx]:
            log.error(f'bad DOI for {i + 2}')
            _r[ridx] = _r[ridx].replace(doibad, 'https://doi.org/')

        for db, rep in doibads:
            if db in _r[ridx]:
                log.error(f'bad citation for {i + 2}')
                _r[ridx] = _r[ridx].replace(db, rep)

        if nogitl in _r[ridx]:
            _coll = []
            for _v in _r[ridx].split(','):
                v = _v.strip()
                if v and nogitl not in v:
                    _coll.append(v)

            _r[ridx] = ','.join(_coll)

        if bad in _r[eidx]:
            log.error(f'bad RRID for {i + 2}')
            _r[eidx] = _r[eidx].replace(bad, 'RRID:SCR_018709')  # FIXME hack
        elif bad2 in _r[eidx]:
            log.error(f'bad RRID for {i + 2}')
            _r[eidx] = _r[eidx].replace(bad2, 'RRID:SCR_018709')  # FIXME hack

        for j, c in enumerate(list(_r)):
            if c.startswith(sigh):
                _r[j] = _r[j].replace(sigh, 'http://uri.interlex.org/')
                log.error(f'bad scheme for uri.interlex.org identifier at {i + 2} {j + 1}')
            elif c.startswith(sigh2):
                _r[j] = _r[j].replace(sigh2, 'http://uri.interlex.org/base/')
                log.error(f'bad prefix for uri.interlex.org identifier at {i + 2} {j + 1}')

    # FIXME TODO maybe skip stuff flagged as invalid
    rows = [[c if c != 'http://www.notspecified.info' else ''
             for c in r] for r in _rows
            if (r[idx] or r[tidx]) and r[sidx]
            and r[idx] != anat_space_hack and r[state_idx] != 'rejected' and r[state_idx] != 'in_progress'
            and r[idx] != expcon and r[oidx] != expcon
            ]
    assert len(rows) > 1

    # missing due to issues with aacar-1 and pancr-2
    aacar_11 = [''] * len(rows[0])
    aacar_11[0] = 'neuron type aacar 11'
    aacar_11[sidx], aacar_11[pidx], aacar_11[idx] = (
        'ilxtr:neuron-type-aacar-11', 'ilxtr:hasForwardConnectionPhenotype', 'ilxtr:neuron-type-aacar-1')

    sdcol_l = [''] * len(rows[0])
    sdcol_l[0] = 'neuron type sdcol l'
    sdcol_l[sidx], sdcol_l[pidx], sdcol_l[idx] = (
        'ilxtr:neuron-type-sdcol-l', 'neurdf.ent:hasMolecularPhenotype', 'TEMPIND:Nos1')

    sdcol_lp = [''] * len(rows[0])
    sdcol_lp[0] = "neuron type sdcol l'"
    sdcol_lp[sidx], sdcol_lp[pidx], sdcol_lp[idx] = (
        'ilxtr:neuron-type-sdcol-l-prime', 'neurdf.ent:hasMolecularPhenotype', 'TEMPIND:Nos1')

    rows.extend((
        aacar_11,
        sdcol_l,
        sdcol_lp,
    ))

    s = sheets.Sheet(fetch=False)
    s.sheet_name = 'comprt'
    s._values = []
    s.update(rows)
    s._uncommitted_appends = {}
    assert len(s.values) > 1
    return s


import os
from pyontutils.core import OntGraph, OntResIri, OntResPath
from neurondm.core import graphBase, OntTerm, OntId, RDFL
from . import apinat_npo  # populate Neuron subclasses
from . import apinat_pops_more  # populate Neuron subclasses
def load_config(gitref='neurons', local=False, restore=True, load_complex=True):
    # FIXME naming ... really load_config_neurons ...
    # FIXME from ../../docs/compoers.py

    config = Config('random-merge')
    g = OntGraph()  # load and query graph

    # remove scigraph and interlex calls
    _old_vocab = graphBase._sgv
    graphBase._sgv = None
    del graphBase._sgv
    if len(OntTerm.query._services) > 1:
        # backup services and avoid issues on rerun
        _old_query_services = OntTerm.query._services
        # noloc excludes the rdflib local service for graphBase.core_graph
        _noloc_query_services = _old_query_services[1:]

    OntTerm.query._services = (RDFL(g, OntId),)

    # base paths to ontology files
    gen_neurons_path = 'ttl/generated/neurons/'
    suffix = '.ttl'
    if local:
        from pyontutils.config import auth
        olr = auth.get_path('ontology-local-repo')
        local_base = olr / gen_neurons_path
    else:
        orr = f'https://raw.githubusercontent.com/SciCrunch/NIF-Ontology/{gitref}/'
        remote_base = orr + gen_neurons_path

    imports = (
        'apinat-partial-orders',
        'apinat-pops-more',
        'apinat-simple-sheet',
        'sparc-nlp',
    )

    if load_complex:
        imports += ('apinat-complex',)

    # full imports
    for f in imports:
        if local:
            ori = OntResPath(local_base / (f + suffix))
        else:
            ori = OntResIri(remote_base + f + suffix)
        [g.add(t) for t in ori.graph]

    # label only imports
    for f in ('apinatomy-neuron-populations',
              '../../npo'):
        p = os.path.normpath(gen_neurons_path + f)
        if local:
            ori = OntResPath(olr / (p + suffix))
        else:
            ori = OntResIri(orr + gen_neurons_path + f + suffix)

        [g.add((s, rdfs.label, o)) for s, o in ori.graph[:rdfs.label:]]

    config.load_existing(g)
    neurons = config.neurons()  # scigraph required here if deps not removed above

    if restore:
        # restore old
        graphBase._sgv = _old_vocab
        OntTerm.query._services = _noloc_query_services

    return neurons, config, g


class NeuronComposer(Neuron):
    owlClass = ilxtr.NeuronComposer
    shortname = 'cmpsr'


def remlabs_write(config, keep=tuple(), extra_trips=tuple()):
    config.write()
    labels = tuple(
        l for l in (#ilxtr.genLabel,
            rdfs.label,
            ilxtr.origLabel,
            ilxtr.localLabel, ilxtr.simpleLabel,
            ilxtr.simpleLocalLabel, skos.prefLabel)
        if l not in keep)
    to_remove = [t for t in config._written_graph if t[1] in labels]
    [config._written_graph.remove(t) for t in to_remove]
    for t in extra_trips:
        config._written_graph.add(t)
    #[config._written_graph.add(t) for t in to_add]
    # TODO figure out about partial orders being roundtripped or not?
    config._written_graph.write()


def ncfun_roundtrip(id):
    if '/neuron-type-' in id:
        return NeuronApinatComplex
    elif '/sparc-nlp/' in id:
        return NeuronSparcNlp
    else:
        return Neuron


def main(report=True):
    exp = aug.LocalPath('~/downloads/export_v6-0-0_2026-01-14_00-57-35.csv').expanduser()

    sht = get_csv_sheet(exp)
    cs = [sht]
    config = Config('composer-and-roundtrip')
    nlp_main(cs=cs, config=config, neuron_class=Neuron, neuron_class_fun=ncfun_roundtrip)  # FIXME neuron_class is incorrect and changes per model
    nrns = config.neurons()
    # FIXME partial order is lost when config changes it seems :/
    # but it sighed was called??? yes, but we add the partial orders
    # in bulk at the end ...

    # TODO
    # use the ids from nrns
    # to create 4 files
    # 1. composer only                -> composer.ttl
    # 2. composer existing only       -> composer-roundtrip.ttl
    # 3. composer existing NOT apinat -> composer-nlp.ttl
    # 4. existing from ontology       -> composer-existing.ttl

    acops = (
        # composer only
        ilxcr.hasComposerURI,

        # composer export derived
        rdfs.label,  # also need remlabs_write keep= for it to propagate
        skos.prefLabel,
        ilxtr.curatorNote,
        #ilxtr.inNLPWorkingSet,  # XXX there is no explicit field in the export for this right now?
        #ilxtr.reference,  # XXX this is not present from composer so leave out to reduce noise
        #ilxtr.reviewNote,  # these cannot be exported right now because the field contains conversations
        #ilxtr.sentenceNumber,
        ilxtr.expertConsultant,

        # ontology derived extra
        #ilxtr.alertNote,
        #ilxtr.hasOrganTarget,  # currently not ingested into composer
        ilxtr.literatureCitation,  # leave this out for now because it is composer only

             )
    def anncop(n, g, include=tuple()):
        s = n.id_
        for p in acops + include:
            for o in g[s:p]:
                graphBase.out_graph.add((s, p, o))

        return n

    from neurondm import orders as nord

    def _key(v):
        if isinstance(v, nord.rl):
            return v
        else:
            return nord.rl(v)

    def pos(nst):
        # XXX temporary hack to normalize all edges for comparison
        _adj = nord.nst_to_adj(nst)
        adj = [tuple(sorted(ab, key=_key)) for ab in _adj]
        return nord.adj_to_nst(adj)

    def pos(nst): return nst

    # skos no longer in rdflib default so not restored after cull_prefixes
    config._written_graph.namespace_manager.bind('skos', str(skos))
    config_composer_only = Config('composer')
    cco_nrns = [
        anncop(NeuronComposer(
            *n.pes, id_=n.id_, label=n._origLabel, override=True, partialOrder=pos(n.partialOrder())),
               config._written_graph)
        for n in nrns if 'sparc-nlp/composer' in n.id_
        or 'composer/uris/set' in n.id_
    ]
    remlabs_write(config_composer_only, keep=(rdfs.label, skos.prefLabel))
    config_composer_only.write_python()

    # FIXME TODO likely need to use OntGraph.subjectGraph to copy these correctly
    config_composer_roundtrip = Config('composer-roundtrip')
    # FIXME TODO use ex_nrns to create a lookup for these
    ccr_nrns = [
        anncop((ncfun_roundtrip(n.id_))(
            *n.pes, id_=n.id_, label=n._origLabel, override=True, partialOrder=pos(n.partialOrder())),
               config._written_graph)
        for n in nrns if 'sparc-nlp/composer' not in n.id_
        and 'composer/uris/set' not in n.id_
    ]
    remlabs_write(config_composer_roundtrip)
    config_composer_roundtrip.write_python()
    rtids = set(n.id_ for n in ccr_nrns)

    config_composer_apinat = Config('apinat-complex')
    cca_nrns = [
        anncop((ncfun_roundtrip(n.id_))(
            *n.pes, id_=n.id_, label=n._origLabel, override=True,
            partialOrder=pos(n.partialOrder())
        ),
               config._written_graph)
        for n in nrns if '/neuron-type-' in n.id_
    ]
    remlabs_write(config_composer_apinat, keep=(rdfs.label, skos.prefLabel))
    config_composer_apinat.write_python()

    # roundtrip without apinat-complex
    config_composer_sheet = Config('composer-nlp')
    ccs_nrns = [
        anncop((ncfun_roundtrip(n.id_))(
            *n.pes, id_=n.id_, label=n._origLabel, override=True, partialOrder=pos(n.partialOrder())),
               config._written_graph, (ilxtr.alertNote,))
        for n in nrns if 'sparc-nlp/composer' not in n.id_
        and 'composer/uris/set' not in n.id_
        and '/neuron-type-' not in n.id_
    ]
    et = ((NeuronSparcNlp.owlClass, rdfs.subClassOf, NeuronComposer.owlClass),)
    remlabs_write(config_composer_sheet, keep=(rdfs.label, skos.prefLabel), extra_trips=et)
    config_composer_sheet.write_python()
    # TODO diff this against sparc-nlp.ttl

    # TODO remove all the gen labels like we do in nlp
    # TODO move ex before composer rt so we can look up the existing parent class
    ex_nrns, ex_config, ex_g = load_config(local=True)
    #ex_nrns = ex_config.neurons()
    config_existing = Config('composer-existing')
    cex_nrns = [
        anncop(n.__class__(
            *n.pes, id_=n.id_, label=n._origLabel, override=True, partialOrder=pos(n.partialOrder())),
               ex_g)
        for n in ex_nrns if n.id_ in rtids]
    remlabs_write(config_existing)
    config_existing.write_python()

    if not report:
        return

    exids = set(n.id_ for n in ex_nrns if '/apinat-simple/' not in n.id_)

    ok_for_reason = {
        ilxtr['neuron-type-pancr-2']: 'apinat contains loops which composer cannot represent, also dendrite edges go in opposite order because composer cannot currently start distally and always starts at soma (see anat space hack)',
        ilxtr['neuron-type-pancr-4']: 'apinat contains a loop which npo skips',
        ilxtr['neuron-type-bolew-unbranched-24']: 'addition of fiber does not change overall order',
    }

    _cnlp_cases = {n.id_: n for n in ccs_nrns}
    _snlp_cases = {n.id_: n for n in ex_nrns if '/sparc-nlp/' in n.id_}
    cnlp_cases = list(_cnlp_cases)
    snlp_cases = list(_snlp_cases)
    ssnlp, scnlp = set(snlp_cases), set(cnlp_cases)
    nlp_miss_cmp = ssnlp - scnlp
    nlp_miss_sht = scnlp - ssnlp
    nlp_cases = list(ssnlp & scnlp)
    nlp_todiff = []
    for c in nlp_cases:
        t = _cnlp_cases[c], _snlp_cases[c]
        nlp_todiff.append(t)

    apinat_cases = [i for i in rtids if 'neuron-type' in i]
    crin = {n.id_:n for n in ccr_nrns}
    exin = {n.id_:n for n in cex_nrns}
    todiff = []
    for a in apinat_cases:
        if a not in crin or a not in exin:
            log.debug(f'one side missing {a}')
            continue

        t = crin[a], exin[a]
        todiff.append(t)

    apn_todiff = todiff

    for i, todiff in enumerate((apn_todiff, nlp_todiff)):
        report_both = {}
        report_cm = {}
        for c, e in todiff:
            _ca = tuple(tuple(e.region if e.layer is None else e for e in ab) for ab in nord.nst_to_adj(c.partialOrder()))
            _ea = nord.nst_to_adj(e.partialOrder())
            ca = tuple(tuple(sorted(p, key=_key)) for p in _ca)
            ea = tuple(tuple(sorted(p, key=_key)) for p in _ea)
            sca, sea = set(ca), set(ea)
            sca_only, sea_only = sca - sea, sea - sca
            # loops
            #sea_maybe_loops = set((b, a) for a, b in sea_only)
            #sea_loops = sca & sea_maybe_loops
            # node diff
            nsca = set(e for es in ca for e in es)
            nsea = set(e for es in ea for e in es)
            nsca_only, nsea_only = nsca - nsea, nsea - nsca
            if sea_only and sca_only:
                msg = f'{c.id_} both different'
                rep = c, e, sca, sea, sca_only, sea_only, nsca_only, nsea_only
                #if len(sea_loops) == len(sea_maybe_loops):
                    #msg = f'{c.id_} all composer missing cases are loops so ok'
                    #log.debug(msg)
                    #continue

                if c.id_ not in ok_for_reason:
                    report_both[c.id_] = rep
            elif sea_only:
                msg = f'{c.id_} composer missing'
                rep = c, e, sca, sea, sca_only, sea_only, nsca_only, nsea_only
                report_cm[c.id_] = rep
            elif sca_only:
                msg = f'{c.id_} only difference is new nodes in composer'
            else:
                msg = None

            if msg:
                log.debug(msg)

        if i == 0:
            apn_report_both = report_both
            apn_report_cm = report_cm
        elif i == 1:
            nlp_report_both = report_both
            nlp_report_cm = report_cm

    ({k:[sorted(_, key=_key) for _ in v[-2:]] for k, v in report_both.items()},
     {k:[sorted(_, key=_key) for _ in v[-2:]] for k, v in report_cm.items()},)

    ({k:[_ for _ in v[-4:-2]] for k, v in report_both.items()},
     {k:[_ for _ in v[-4:-2]] for k, v in report_cm.items()},)

    {k:[_ for _ in v[-4:-2]] for k, v in nlp_report_both.items()}
    {k:[sorted(_, key=_key) for _ in v[-2:]] for k, v in nlp_report_both.items() if v[-2] or v[-1]}

    breakpoint()
    return config,


def _oldmain():
    r1 = sht.rows()[1]
    trips = [t for t in [row_to_triple(r) for r in sht.rows()[1:]] if t is not None]
    bund = {}
    for s, p, o in trips:
        if s not in bund:
            bund[s] = []

        bund[s].append((p, o))

def row_to_triple(row):  # XXX old before reuse of nlp code
    skip = ilxtr.hasSomaPhenotype,
    s = rdflib.URIRef(row.uri().value)

    p = map_predicates(row.predicate().value)
    if p in skip:
        return
        
    v = row.identifier().value
    if v:
        if ',' in v:
            r, l = v.split(',')
            o = rdflib.URIRef(r)  # TODO region + layer
        else:
            o = rdflib.URIRef(v)
    else:
        # looks like the export contains rows even when there is no value for that phenotype
        # which is annoying because you can't tell if you screwed up or whether the data
        # actually did not contain that value
        msg = f'where value for {p}?'
        #raise NotImplementedError(msg)
        log.error(msg)
        o = TEMP.FIXME_MISSING
        return

    row.connected_from_uris()  # used for partial order probably
    row.axonal_course_poset()
    #row.relationship()
    #s.rows()[1].cells[0]
    #s.raw_values = rows
    #s._values = rows
    return s, p, o


if __name__ == '__main__':
    main()
